from .deterministic import _compute_profit_credit_scoring, mpcs, mpcs_score
from .stochastic import empcs, empcs_score

__all__ = [
    '_compute_profit_credit_scoring',
    'empcs',
    'empcs_score',
    'mpcs',
    'mpcs_score',
]



import sys
import warnings
from collections.abc import Callable
from numbers import Real
from typing import Any, ClassVar, Literal

import numpy as np
from numpy.typing import ArrayLike
from scipy.optimize import OptimizeResult, minimize
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils._param_validation import StrOptions

from ..._common import Parameter
from ..._types import FloatArrayLike, FloatNDArray, IntNDArray, ParameterConstraint
from ...metrics import Metric
from ...metrics.metric.prebuilt_metrics import make_generic_cost_metric
from .._base import BaseLogitClassifier, OptimizeFn
from ._cs_mixin import CostSensitiveMixin

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
LossStr = Literal['average expected cost']
GradientLossFn = Callable[[FloatNDArray, FloatNDArray, FloatNDArray], tuple[float, FloatNDArray]]
ObjectiveFn = Callable[..., float | tuple[float, FloatNDArray] | tuple[float, FloatNDArray, FloatNDArray]]


class CSLogitClassifier(BaseLogitClassifier, CostSensitiveMixin):
    """
    Cost-sensitive logistic regression classifier.

    Read more in the :ref:`User Guide <cslogit>`.

    .. seealso::

        :func:`~empulse.metrics.make_objective_aec` : Creates the instance-dependent cost function.

        :class:`~empulse.models.CSBoostClassifier` : Cost-sensitive XGBoost classifier.

    Parameters
    ----------
    C : float, default=1.0
        Inverse of regularization strength; must be a positive ``float``.
        Like in support vector machines, smaller values specify stronger regularization.

    fit_intercept : bool, default=True
        Specifies if a constant (a.k.a. bias or intercept) should be added to the decision function.

    soft_threshold : bool, default=False
        If ``True``, apply soft-thresholding to the regression coefficients.

    l1_ratio : float, default=1.0
        The ElasticNet mixing parameter, with ``0 <= l1_ratio <= 1``.

            - For ``l1_ratio = 0`` the penalty is a L2 penalty.
            - For ``l1_ratio = 1`` it is a L1 penalty.
            - For ``0 < l1_ratio < 1``, the penalty is a combination of L1 and L2.

    loss : {'average expected cost'}, Callable or :class:`empulse.metrics.Metric`, default='average expected cost'
        Loss function which should be minimized.

        - If ``str``, then it should be one of the following:

            - 'average expected cost' : Average Expected Cost loss function,
              see :func:`~empulse.metrics.expected_cost_loss`.

        - If ``Callable`` it should have a signature ``loss(y_true, y_score)``.

        - If :class`~empulse.metrics.Metric`, metric parameters are passed as ``loss_params``
          to the :Meth:`~empulse.models.CSLogitClassifier.fit` method.

        By default, loss function is minimized, customize behaviour in `optimize_fn`.

    optimize_fn : Callable, optional
        Optimization algorithm. Should be a Callable with signature ``optimize(objective, X)``.
        See :ref:`proflogit` for more information.

    optimizer_params : dict[str, Any], optional
        Additional keyword arguments passed to `optimize_fn`.

        tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    tn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true negatives. If ``float``, then all true negatives have the same cost.
        If array-like, then it is the cost of each true negative classification.
        Is overwritten if another `tn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    Attributes
    ----------
    classes_ : numpy.ndarray
        Unique classes in the target found during fit.

    result_ : :class:`scipy:scipy.optimize.OptimizeResult`
        Optimization result.

    coef_ : numpy.ndarray, shape=(n_features,)
        Coefficients of the logit model.

    intercept_ : float
        Intercept of the logit model.
        Only available when ``fit_intercept=True``.

    Examples
    --------

    .. code-block:: python

        import numpy as np
        from empulse.models import CSLogitClassifier
        from sklearn.datasets import make_classification

        X, y = make_classification()
        fn_cost = np.random.rand(y.size)  # instance-dependent cost
        fp_cost = 5  # constant cost

        model = CSLogitClassifier(C=0.1)
        model.fit(X, y, fn_cost=fn_cost, fp_cost=fp_cost)
        y_proba = model.predict_proba(X)

    Example with passing instance-dependent costs through cross-validation:

    .. code-block:: python

        import numpy as np
        from empulse.models import CSLogitClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import cross_val_score
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        fn_cost = np.random.rand(y.size)
        fp_cost = 5

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('model', CSLogitClassifier(C=0.1).set_fit_request(fn_cost=True, fp_cost=True)),
        ])

        cross_val_score(pipeline, X, y, params={'fn_cost': fn_cost, 'fp_cost': fp_cost})

    Example with passing instance-dependent costs through a grid search:

    .. code-block:: python

        import numpy as np
        from empulse.metrics import expected_cost_loss
        from empulse.models import CSLogitClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import GridSearchCV
        from sklearn.metrics import make_scorer
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler

        set_config(enable_metadata_routing=True)

        X, y = make_classification(n_samples=50)
        fn_cost = np.random.rand(y.size)
        fp_cost = 5

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('model', CSLogitClassifier().set_fit_request(fn_cost=True, fp_cost=True)),
        ])
        param_grid = {'model__C': np.logspace(-5, 2, 5)}
        scorer = make_scorer(
            expected_cost_loss,
            response_method='predict_proba',
            greater_is_better=False,
            normalize=True,
        )
        scorer = scorer.set_score_request(fn_cost=True, fp_cost=True)

        grid_search = GridSearchCV(pipeline, param_grid=param_grid, scoring=scorer)
        grid_search.fit(X, y, fn_cost=fn_cost, fp_cost=fp_cost)

    References
    ----------
    .. [1] Höppner, S., Baesens, B., Verbeke, W., & Verdonck, T. (2022).
           Instance-dependent cost-sensitive learning for detecting transfer fraud.
           European Journal of Operational Research, 297(1), 291-300.
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        **BaseLogitClassifier._parameter_constraints,
        'loss': [StrOptions({'average expected cost'}), Metric],
        'tp_cost': ['array-like', Real],
        'tn_cost': ['array-like', Real],
        'fn_cost': ['array-like', Real],
        'fp_cost': ['array-like', Real],
    }

    def __init__(
        self,
        *,
        C: float = 1.0,
        fit_intercept: bool = True,
        soft_threshold: bool = False,
        l1_ratio: float = 1.0,
        loss: LossStr | Metric = 'average expected cost',
        optimize_fn: OptimizeFn | None = None,
        optimizer_params: dict[str, Any] | None = None,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
    ):
        super().__init__(
            C=C,
            fit_intercept=fit_intercept,
            soft_threshold=soft_threshold,
            l1_ratio=l1_ratio,
            loss=loss,
            optimize_fn=optimize_fn,
            optimizer_params=optimizer_params,
        )
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fn_cost = fn_cost
        self.fp_cost = fp_cost

    def _get_metric_loss(self) -> Metric | None:
        """Get the metric loss function if available."""
        if isinstance(self.loss, Metric):
            return self.loss
        return None

    def fit(
        self,
        X: FloatArrayLike,
        y: ArrayLike,
        *,
        tp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        tn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        **loss_params: Any,
    ) -> Self:
        """
        Fit the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        y : array-like of shape (n_samples,)

        tp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true positives. If ``float``, then all true positives have the same cost.
            If array-like, then it is the cost of each true positive classification.

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        tn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true negatives. If ``float``, then all true negatives have the same cost.
            If array-like, then it is the cost of each true negative classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        loss_params : dict[str, Any]
            Additional parameters passed to the loss function.

        Returns
        -------
        self : CSLogitClassifier
            Fitted model.

        """
        super().fit(X, y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, **loss_params)
        return self

    def _fit(
        self,
        X: FloatNDArray,
        y: IntNDArray,
        *,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
        **loss_params: Any,
    ) -> Self:
        optimizer_params = self.optimizer_params or {}

        if not isinstance(self.loss, Metric):
            tp_cost, tn_cost, fn_cost, fp_cost = self._check_costs(
                tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost
            )

            if not isinstance(tp_cost, Real) and (tp_cost := np.asarray(tp_cost)).ndim == 1:
                tp_cost = np.expand_dims(tp_cost, axis=1)
            if not isinstance(tn_cost, Real) and (tn_cost := np.asarray(tn_cost)).ndim == 1:
                tn_cost = np.expand_dims(tn_cost, axis=1)
            if not isinstance(fn_cost, Real) and (fn_cost := np.asarray(fn_cost)).ndim == 1:
                fn_cost = np.expand_dims(fn_cost, axis=1)
            if not isinstance(fp_cost, Real) and (fp_cost := np.asarray(fp_cost)).ndim == 1:
                fp_cost = np.expand_dims(fp_cost, axis=1)

            # Assume that the loss function takes the following parameters:
            loss_params['tp_cost'] = tp_cost
            loss_params['tn_cost'] = tn_cost
            loss_params['fn_cost'] = fn_cost
            loss_params['fp_cost'] = fp_cost

        loss = self.loss if isinstance(self.loss, Metric) else make_generic_cost_metric()
        objective = loss._logit_objective(
            features=X,
            y_true=y,
            C=self.C,
            l1_ratio=self.l1_ratio,
            soft_threshold=self.soft_threshold,
            fit_intercept=self.fit_intercept,
            **loss_params,
        )
        optimize_fn: Callable[..., OptimizeResult] = (
            _optimize_jacobian if self.optimize_fn is None else self.optimize_fn
        )  # type: ignore[no-redef]
        self.result_ = optimize_fn(objective=objective, X=X, **optimizer_params)
        self.coef_ = self.result_.x[1:] if self.fit_intercept else self.result_.x
        if self.fit_intercept:
            self.intercept_ = self.result_.x[0]
        return self


def _optimize_jacobian(
    objective: Callable[[FloatNDArray], tuple[float, FloatNDArray]],
    X: FloatNDArray,
    max_iter: int = 1000,
    tolerance: float = 1e-4,
    **kwargs: Any,
) -> OptimizeResult:
    initial_weights = np.zeros(X.shape[1], order='F', dtype=X.dtype)

    result = minimize(
        objective,
        initial_weights,
        method='L-BFGS-B',
        jac=True,
        options={
            'maxiter': max_iter,
            'maxls': 50,
            'gtol': tolerance,
            'ftol': 64 * np.finfo(float).eps,
        },
        **kwargs,
    )
    _check_optimize_result(result)

    return result


def _check_optimize_result(result: OptimizeResult) -> None:
    """
    Check the OptimizeResult for successful convergence.

    Parameters
    ----------
    result : OptimizeResult
       Result of the scipy.optimize.minimize function.
    """
    # handle both scipy and scikit-learn solver names
    if result.status != 0:
        try:
            # The message is already decoded in scipy>=1.6.0
            result_message = result.message.decode('latin1')
        except AttributeError:
            result_message = result.message
        warning_msg = (
            f'L-BFGS failed to converge (status={result.status}):\n{result_message}.\n\n'
            'Increase the number of iterations (max_iter) '
            'or scale the data as shown in:\n'
            '    https://scikit-learn.org/stable/modules/'
            'preprocessing.html'
        )
        warnings.warn(warning_msg, ConvergenceWarning, stacklevel=2)



import numpy as np
from numpy.typing import ArrayLike

from ._convex_hull import _compute_convex_hull


def max_profit_score(
    y_true: ArrayLike,
    y_score: ArrayLike,
    *,
    tp_benefit: float = 0.0,
    tn_benefit: float = 0.0,
    fn_cost: float = 0.0,
    fp_cost: float = 0.0,
) -> float:
    """
    :func:`~empulse.metrics.max_profit()` but only returning the MP score.

    .. seealso::
        :func:`~empulse.metrics.max_profit` : To also return the threshold at which the maximum profit is achieved.

        :func:`~empulse.metrics.Metric` : To create a stochastic version of the maximum profit.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('positive': 1, 'negative': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    tp_benefit : float, default=0.0
        Benefit attributed to true positive predictions.

    tn_benefit : float, default=0.0
        Benefit attributed to true negative predictions.

    fn_cost : float, default=0.0
        Cost attributed to false negative predictions.

    fp_cost : float, default=0.0
        Cost attributed to false positive predictions.

    Returns
    -------
    mp : float
        Maximum Profit.

    Notes
    -----
    The MP is defined as [1]_:

    .. math::

        \\text{MP} = b_0 \\pi_0 F_0(T) + b_1 \\pi_1 (1 - F_1(T)) - c_0 \\pi_0 (1 - F_0(T)) - c_1 F_1(T)

    where :math:`T` is the threshold at which the maximum profit is achieved.

    The MP requires that the positive class is encoded as 0, and negative class as 1.
    However, this implementation assumes the standard notation ('positive': 1, 'negative': 0).

    References
    ----------
    .. [1] Verbraken, T., Verbeke, W. and Baesens, B. (2013).
        A Novel Profit Maximizing Metric for Measuring Classification
        Performance of Customer Churn Prediction Models. IEEE Transactions on
        Knowledge and Data Engineering, 25(5), 961-973. Available Online:
        http://ieeexplore.ieee.org/iel5/69/6486492/06165289.pdf?arnumber=6165289

    Examples
    --------
    Reimplement MPC:

    >>> from empulse.metrics import max_profit_score
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>>
    >>> clv = 200
    >>> d = 10
    >>> f = 1
    >>> gamma = 0.3
    >>> tp_benefit = clv * (gamma * (1 - (d / clv) - (f / clv)))
    >>> fp_cost = d + f
    >>>
    >>> max_profit_score(y_true, y_score, tp_benefit=tp_benefit, fp_cost=fp_cost)
    24.22...
    """
    return max_profit(y_true, y_score, tp_benefit=tp_benefit, tn_benefit=tn_benefit, fn_cost=fn_cost, fp_cost=fp_cost)[
        0
    ]


def max_profit(
    y_true: ArrayLike,
    y_score: ArrayLike,
    *,
    tp_benefit: float = 0.0,
    tn_benefit: float = 0.0,
    fn_cost: float = 0.0,
    fp_cost: float = 0.0,
) -> tuple[float, float]:
    """
    Maximum Profit Measure (MP).

    .. seealso::
        :func:`~empulse.metrics.max_profit_score` : To only return the maximum profit score.

        :func:`~empulse.metrics.Metric` : To create a stochastic version of the maximum profit.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('positive': 1, 'negative': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    tp_benefit : float, default=0.0
        Benefit attributed to true positive predictions.

    tn_benefit : float, default=0.0
        Benefit attributed to true negative predictions.

    fn_cost : float, default=0.0
        Cost attributed to false negative predictions.

    fp_cost : float, default=0.0
        Cost attributed to false positive predictions.

    Returns
    -------
    mp : float
        Maximum Profit

    threshold : float
        Threshold at which the maximum profit is achieved

    Notes
    -----
    The MP is defined as [1]_:

    .. math::

        \\text{MP} = b_0 \\pi_0 F_0(T) + b_1 \\pi_1 (1 - F_1(T)) - c_0 \\pi_0 (1 - F_0(T)) - c_1 F_1(T)

    where :math:`T` is the threshold at which the maximum profit is achieved.

    The MP requires that the positive class is encoded as 0, and negative class as 1.
    However, this implementation assumes the standard notation ('positive': 1, 'negative': 0).

    References
    ----------
    .. [1] Verbraken, T., Verbeke, W. and Baesens, B. (2013).
        A Novel Profit Maximizing Metric for Measuring Classification
        Performance of Customer Churn Prediction Models. IEEE Transactions on
        Knowledge and Data Engineering, 25(5), 961-973. Available Online:
        http://ieeexplore.ieee.org/iel5/69/6486492/06165289.pdf?arnumber=6165289

    Examples
    --------
    Reimplement MPC:

    >>> from empulse.metrics import max_profit
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>>
    >>> clv = 200
    >>> d = 10
    >>> f = 1
    >>> gamma = 0.3
    >>> tp_benefit = clv * (gamma * (1 - (d / clv)) - (f / clv))
    >>> fp_cost = d + f
    >>>
    >>> max_profit(y_true, y_score, tp_benefit=tp_benefit, fp_cost=fp_cost)
    (23.87..., 0.875)
    """
    y_true = np.asarray(y_true)
    y_score = np.asarray(y_score)

    pi0 = float(np.mean(y_true))
    pi1 = 1 - pi0

    f0, f1 = _compute_convex_hull(y_true, y_score)

    profits = (tp_benefit + fn_cost) * pi0 * f0 - (tn_benefit + fp_cost) * pi1 * f1 + tn_benefit * pi1 - fn_cost * pi0
    best_index = np.argmax(profits)
    maximum_profit = float(profits[best_index])
    customer_threshold = float(f0[best_index] * pi0 + f1[best_index] * pi1)
    return maximum_profit, customer_threshold



import sys
from collections.abc import Callable
from functools import partial
from typing import Any, ClassVar, Literal

import numpy as np
import sympy

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ..._types import FloatNDArray
from .._loss import cy_logit_loss_gradient
from .common import (
    BoostGradientConst,
    Direction,
    LogitConsts,
    MetricFn,
    RateFn,
    SympyFnPickleMixin,
    ThresholdFn,
    _check_parameters,
    _safe_lambdify,
    _safe_run_lambda,
)
from .cost_metric import (
    CostBoostGradientConst,
    CostLogitConsts,
    CostLoss,
    CostOptimalRate,
    CostOptimalThreshold,
    _build_cost_equation,
    _format_cost_function,
)
from .metric_strategies import MetricStrategy


class Savings(MetricStrategy):
    """Strategy for the Expected Savings metric."""

    def __init__(self) -> None:
        super().__init__(name='savings', direction=Direction.MAXIMIZE)

    def build(
        self,
        tp_benefit: sympy.Expr,
        tn_benefit: sympy.Expr,
        fp_cost: sympy.Expr,
        fn_cost: sympy.Expr,
    ) -> Self:
        """Build the metric strategy."""
        self._score_function: MetricFn = SavingsScore(  # type: ignore[assignment]
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        self._optimal_threshold: ThresholdFn = CostOptimalThreshold(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        self._optimal_rate: RateFn = CostOptimalRate(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        self._score_logit_function = CostLoss(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        self._prepare_logit_objective: LogitConsts = CostLogitConsts(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        self._prepare_boost_objective: BoostGradientConst = CostBoostGradientConst(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        return self

    def score(  # type: ignore[override]
        self,
        y_true: FloatNDArray,
        y_score: FloatNDArray,
        baseline: Literal['zero_one', 'zero', 'one', 'prior'] | FloatNDArray = 'zero_one',
        **parameters: FloatNDArray | float,
    ) -> float:
        """
        Compute the metric expected savings score.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        score: float
            The expected savings score.
        """
        return self._score_function(y_true, y_score, baseline=baseline, **parameters)

    def optimal_threshold(
        self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: FloatNDArray | float
    ) -> float | FloatNDArray:
        """
        Compute the classification threshold(s) to optimize the metric value.

        i.e., the score threshold at which an observation should be classified as positive to optimize the metric.
        For instance-dependent costs and benefits, this will return an array of thresholds, one for each sample.
        For class-dependent costs and benefits, this will return a single threshold value.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        optimal_threshold: float | FloatNDArray
            The optimal classification threshold(s).
        """
        return self._optimal_threshold(y_true, y_score, **parameters)

    def optimal_rate(self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: FloatNDArray | float) -> float:
        """
        Compute the predicted positive rate to optimize the metric value.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        optimal_rate: float
            The optimal predicted positive rate.
        """
        return self._optimal_rate(y_true, y_score, **parameters)

    def prepare_logit_objective(
        self, features: FloatNDArray, y_true: FloatNDArray, **parameters: FloatNDArray | float
    ) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray]:
        """
        Compute the constant term of the loss and gradient of the metric wrt logistic regression coefficients.

        Parameters
        ----------
        features : NDArray of shape (n_samples, n_features)
            The features of the samples.
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        gradient_const : NDArray of shape (n_samples, n_features)
            The constant term of the gradient.
        loss_const1 : NDArray of shape (n_features,)
            The first constant term of the loss function.
        loss_const2 : NDArray of shape (n_features,)
            The second constant term of the loss function.
        """
        if y_true.ndim == 1:
            y_true = np.expand_dims(y_true, axis=1)
        return self._prepare_logit_objective.prepare(features, y_true, **parameters)

    def logit_objective(
        self,
        features: FloatNDArray,
        y_true: FloatNDArray,
        C: float,
        l1_ratio: float,
        soft_threshold: bool,
        fit_intercept: bool,
        **parameters: FloatNDArray | float,
    ) -> Callable[[FloatNDArray], tuple[float, FloatNDArray]]:
        """
        Build a function which computes the metric value and the gradient of the metric w.r.t logistic coefficients.

        Parameters
        ----------
        features : NDArray of shape (n_samples, n_features)
            The features of the samples.
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        C : float
            Regularization strength parameter. Smaller values specify stronger regularization.
        l1_ratio : float
            The Elastic-Net mixing parameter, with range 0 <= l1_ratio <= 1.
            l1_ratio=0 corresponds to L2 penalty, l1_ratio=1 to L1 penalty.
        soft_threshold : bool
            Indicator of whether soft thresholding is applied during optimization.
        fit_intercept : bool
            Specifies if an intercept should be included in the model.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        logistic_objective : Callable[[NDArray], tuple[float, NDArray]]
            A function that takes logistic regression weights as input and returns the metric value and its gradient.
            The function signature is:
            ``logistic_objective(weights) -> (value, gradient)``
        """
        grad_const, loss_const1, loss_const2 = self.prepare_logit_objective(features, y_true, **parameters)
        loss_const1 = (
            loss_const1.reshape(-1)
            if isinstance(loss_const1, np.ndarray)
            else np.full(len(y_true), loss_const1, dtype=np.float64)
        )
        loss_const2 = (
            loss_const2.reshape(-1)
            if isinstance(loss_const2, np.ndarray)
            else np.full(len(y_true), loss_const2, dtype=np.float64)
        )
        return partial(
            cy_logit_loss_gradient,
            grad_const=grad_const,
            loss_const1=loss_const1,
            loss_const2=loss_const2,
            features=features,
            C=C,
            l1_ratio=l1_ratio,
            soft_threshold=soft_threshold,
            fit_intercept=fit_intercept,
        )

    def prepare_boost_objective(self, y_true: FloatNDArray, **parameters: FloatNDArray | float) -> FloatNDArray:
        """
        Compute the gradient's constant term of the metric wrt gradient boost.

        Parameters
        ----------
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        gradient_const : NDArray of shape (n_samples, n_features)
            The constant term of the gradient.
        """
        if y_true.ndim == 1:
            y_true = np.expand_dims(y_true, axis=1)
        return self._prepare_boost_objective(y_true, **parameters)

    def to_latex(
        self,
        tp_benefit: sympy.Expr,
        tn_benefit: sympy.Expr,
        fp_cost: sympy.Expr,
        fn_cost: sympy.Expr,
    ) -> str:
        """Return the LaTeX representation of the metric."""
        return _savings_score_to_latex(tp_benefit, tn_benefit, fp_cost, fn_cost)


class SavingsScore(SympyFnPickleMixin):
    """Class to compute the metric for binary classification."""

    _sympy_functions: ClassVar[dict[str, str]] = {'cost_function': 'cost_equation'}

    def __init__(self, tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr):
        self.cost_equation = _build_cost_equation(
            tp_cost=-tp_benefit, tn_cost=-tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost
        )
        if any(sympy.stats.rv.is_random(symbol) for symbol in self.cost_equation.free_symbols):
            raise NotImplementedError('Random variables are not supported for the savings metric.')
        self.all_zero_equation, self.all_one_equation = _build_naive_cost_functions(self.cost_equation)

        self.cost_func = _safe_lambdify(self.cost_equation)
        self.all_zero_function = _safe_lambdify(self.all_zero_equation)
        self.all_one_function = _safe_lambdify(self.all_one_equation)

    def __call__(
        self,
        y_true: FloatNDArray,
        y_score: FloatNDArray,
        baseline: FloatNDArray | Literal['zero_one', 'zero', 'one', 'prior'],
        **kwargs: Any,
    ) -> float:
        """Compute the savings score."""
        all_symbols = (
            self.cost_equation.free_symbols | self.all_zero_equation.free_symbols | self.all_one_equation.free_symbols
        )
        _check_parameters(all_symbols - {*sympy.symbols('y s')}, kwargs)

        if isinstance(baseline, np.ndarray):
            cost_base = float(
                np.mean(_safe_run_lambda(self.cost_func, self.cost_equation, y=y_true, s=baseline, **kwargs))
            )
        elif baseline == 'zero_one':
            all_zero_score = float(
                np.mean(_safe_run_lambda(self.all_zero_function, self.all_zero_equation, y=y_true, **kwargs))
            )
            all_one_score = float(
                np.mean(_safe_run_lambda(self.all_one_function, self.all_one_equation, y=y_true, **kwargs))
            )
            cost_base = min(all_zero_score, all_one_score)
        elif baseline == 'zero':
            cost_base = float(
                np.mean(_safe_run_lambda(self.all_zero_function, self.all_zero_equation, y=y_true, **kwargs))
            )
        elif baseline == 'one':
            cost_base = float(
                np.mean(_safe_run_lambda(self.all_one_function, self.all_one_equation, y=y_true, **kwargs))
            )
        elif baseline == 'prior':
            prior = np.mean(y_true)
            cost_base = float(
                np.mean(
                    _safe_run_lambda(
                        self.cost_func, self.cost_equation, y=y_true, s=np.full_like(y_true, prior), **kwargs
                    )
                )
            )
        else:
            raise ValueError("Invalid baseline. Must be 'zero_one', 'zero', 'one', 'prior', or an array-like.")

        if cost_base == 0.0:
            cost_base = float(np.finfo(float).eps)
        cost = _safe_run_lambda(self.cost_func, self.cost_equation, y=y_true, s=y_score, **kwargs)
        return float(1 - np.mean(cost) / cost_base)


def _build_naive_cost_functions(cost_function: sympy.Expr) -> tuple[sympy.Expr, sympy.Expr]:
    all_zero_function = cost_function.subs('s', 0)
    all_one_function = cost_function.subs('s', 1)
    return all_zero_function, all_one_function


def _savings_score_to_latex(
    tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr
) -> str:
    from sympy.printing.latex import latex

    i, N, c0, c1 = sympy.symbols('i N Cost_{0} Cost_{1}')  # noqa: N806
    savings_function = (1 / (N * sympy.Min(c0, c1))) * sympy.Sum(
        _format_cost_function(tp_cost=-tp_benefit, tn_cost=-tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost), (i, 0, N)
    )

    for symbol in savings_function.free_symbols:
        if symbol not in {N, c0, c1}:
            savings_function = savings_function.subs(symbol, str(symbol) + '_i')

    output = latex(savings_function, mode='plain', order=None)

    return f'$\\displaystyle {output}$'



import sys
from collections.abc import Callable
from functools import partial
from typing import Any, ClassVar

import numpy as np
import sympy

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ..._types import FloatNDArray
from .._loss import cy_logit_loss_gradient
from .common import (
    BoostGradientConst,
    Direction,
    LogitConsts,
    MetricFn,
    RateFn,
    SympyFnPickleMixin,
    ThresholdFn,
    _check_parameters,
    _safe_lambdify,
    _safe_run_lambda,
    _safe_run_lambda_array,
)
from .metric_strategies import MetricStrategy


class Cost(MetricStrategy):
    """Strategy for the Expected Cost metric."""

    def __init__(self) -> None:
        super().__init__(name='cost', direction=Direction.MINIMIZE)

    def build(
        self,
        tp_benefit: sympy.Expr,
        tn_benefit: sympy.Expr,
        fp_cost: sympy.Expr,
        fn_cost: sympy.Expr,
    ) -> Self:
        """Build the metric strategy."""
        all_symbols = tp_benefit.free_symbols | tn_benefit.free_symbols | fp_cost.free_symbols | fn_cost.free_symbols
        if any(sympy.stats.rv.is_random(symbol) for symbol in all_symbols):
            raise NotImplementedError('Random variables are not supported for the cost metric.')

        self._score_function: MetricFn = CostLoss(
            tp_benefit=tp_benefit, tn_benefit=tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost
        )
        self._optimal_threshold: ThresholdFn = CostOptimalThreshold(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        self._optimal_rate: RateFn = CostOptimalRate(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        self._prepare_logit_objective: LogitConsts = CostLogitConsts(
            tp_benefit=tp_benefit, tn_benefit=tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost
        )
        self._prepare_boost_objective: BoostGradientConst = CostBoostGradientConst(
            tp_benefit=tp_benefit,
            tn_benefit=tn_benefit,
            fp_cost=fp_cost,
            fn_cost=fn_cost,
        )
        return self

    def score(self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: FloatNDArray | float) -> float:
        """
        Compute the metric expected cost loss.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        score: float
            The expected cost loss.
        """
        return self._score_function(y_true, y_score, **parameters)

    def optimal_threshold(
        self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: FloatNDArray | float
    ) -> float | FloatNDArray:
        """
        Compute the classification threshold(s) to optimize the metric value.

        i.e., the score threshold at which an observation should be classified as positive to optimize the metric.
        For instance-dependent costs and benefits, this will return an array of thresholds, one for each sample.
        For class-dependent costs and benefits, this will return a single threshold value.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        optimal_threshold: float | FloatNDArray
            The optimal classification threshold(s).
        """
        return self._optimal_threshold(y_true, y_score, **parameters)

    def optimal_rate(self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: FloatNDArray | float) -> float:
        """
        Compute the predicted positive rate to optimize the metric value.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        optimal_rate: float
            The optimal predicted positive rate.
        """
        return self._optimal_rate(y_true, y_score, **parameters)

    def prepare_logit_objective(
        self, features: FloatNDArray, y_true: FloatNDArray, **parameters: FloatNDArray | float
    ) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray]:
        """
        Compute the constant term of the loss and gradient of the metric wrt logistic regression coefficients.

        Parameters
        ----------
        features : NDArray of shape (n_samples, n_features)
            The features of the samples.
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        gradient_const : NDArray of shape (n_samples, n_features)
            The constant term of the gradient.
        loss_const1 : NDArray of shape (n_features,)
            The first constant term of the loss function.
        loss_const2 : NDArray of shape (n_features,)
            The second constant term of the loss function.
        """
        if y_true.ndim == 1:
            y_true = np.expand_dims(y_true, axis=1)

        return self._prepare_logit_objective.prepare(features, y_true, **parameters)

    def logit_objective(
        self,
        features: FloatNDArray,
        y_true: FloatNDArray,
        C: float,
        l1_ratio: float,
        soft_threshold: bool,
        fit_intercept: bool,
        **parameters: FloatNDArray | float,
    ) -> Callable[[FloatNDArray], tuple[float, FloatNDArray]]:
        """
        Build a function which computes the metric value and the gradient of the metric w.r.t logistic coefficients.

        Parameters
        ----------
        features : NDArray of shape (n_samples, n_features)
            The features of the samples.
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        C : float
            Regularization strength parameter. Smaller values specify stronger regularization.
        l1_ratio : float
            The Elastic-Net mixing parameter, with range 0 <= l1_ratio <= 1.
            l1_ratio=0 corresponds to L2 penalty, l1_ratio=1 to L1 penalty.
        soft_threshold : bool
            Indicator of whether soft thresholding is applied during optimization.
        fit_intercept : bool
            Specifies if an intercept should be included in the model.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        logistic_objective : Callable[[NDArray], tuple[float, NDArray]]
            A function that takes logistic regression weights as input and returns the metric value and its gradient.
            The function signature is:
            ``logistic_objective(weights) -> (value, gradient)``
        """
        grad_const, loss_const1, loss_const2 = self.prepare_logit_objective(features, y_true, **parameters)
        loss_const1 = (
            loss_const1.reshape(-1)
            if isinstance(loss_const1, np.ndarray)
            else np.full(len(y_true), loss_const1, dtype=np.float64)
        )
        loss_const2 = (
            loss_const2.reshape(-1)
            if isinstance(loss_const2, np.ndarray)
            else np.full(len(y_true), loss_const2, dtype=np.float64)
        )
        return partial(
            cy_logit_loss_gradient,
            grad_const=grad_const,
            loss_const1=loss_const1,
            loss_const2=loss_const2,
            features=features,
            C=C,
            l1_ratio=l1_ratio,
            soft_threshold=soft_threshold,
            fit_intercept=fit_intercept,
        )

    def prepare_boost_objective(self, y_true: FloatNDArray, **parameters: FloatNDArray | float) -> FloatNDArray:
        """
        Compute the gradient's constant term of the metric wrt gradient boost.

        Parameters
        ----------
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        gradient_const : NDArray of shape (n_samples, n_features)
            The constant term of the gradient.
        """
        if y_true.ndim == 1:
            y_true = np.expand_dims(y_true, axis=1)
        return self._prepare_boost_objective(y_true, **parameters)

    def to_latex(
        self,
        tp_benefit: sympy.Expr,
        tn_benefit: sympy.Expr,
        fp_cost: sympy.Expr,
        fn_cost: sympy.Expr,
    ) -> str:
        """Return the LaTeX representation of the metric."""
        return _cost_loss_to_latex(tp_benefit, tn_benefit, fp_cost, fn_cost)


class CostLoss(SympyFnPickleMixin):
    """Class to compute the metric for binary classification."""

    _sympy_functions: ClassVar[dict[str, str]] = {'cost_function': 'cost_equation'}

    def __init__(self, tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr):
        self.cost_equation = _build_cost_equation(
            tp_cost=-tp_benefit, tn_cost=-tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost
        )
        self.cost_function = _safe_lambdify(self.cost_equation)

    def __call__(self, y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        """Compute the cost loss."""
        _check_parameters(self.cost_equation.free_symbols - set(sympy.symbols('y s')), kwargs)
        return float(np.mean(_safe_run_lambda(self.cost_function, self.cost_equation, y=y_true, s=y_score, **kwargs)))


def _build_cost_equation(
    tp_cost: sympy.Expr, tn_cost: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr
) -> sympy.Expr:
    y, s = sympy.symbols('y s')
    cost_function = y * (s * tp_cost + (1 - s) * fn_cost) + (1 - y) * ((1 - s) * tn_cost + s * fp_cost)
    return cost_function


class CostLogitConsts(SympyFnPickleMixin):
    """Class to compute the constants of the cost metric for logistic regression."""

    _sympy_functions: ClassVar[dict[str, str]] = {
        'gradient_fn': 'gradient_const',
        'loss_const1_fn': 'loss_const1',
        'loss_const2_fn': 'loss_const2',
    }

    def __init__(self, tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr):
        y, x = sympy.symbols('y x')
        self.gradient_const = x * (y * (-tp_benefit - fn_cost) + (1 - y) * (fp_cost + tn_benefit))
        self.gradient_fn = _safe_lambdify(self.gradient_const)
        self.loss_const1 = y * -tp_benefit + (1 - y) * fp_cost
        self.loss_const1_fn = _safe_lambdify(self.loss_const1)
        self.loss_const2 = y * fn_cost - (1 - y) * tn_benefit
        self.loss_const2_fn = _safe_lambdify(self.loss_const2)

    def prepare(
        self, x: FloatNDArray, y_true: FloatNDArray, **kwargs: Any
    ) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray]:
        """Prepare the constant terms for the logistic regression objective."""
        gradient_const_value = _safe_run_lambda_array(
            self.gradient_fn, self.gradient_const, shape=x.shape, y=y_true, x=x, **kwargs
        )
        loss_const1_value = _safe_run_lambda_array(
            self.loss_const1_fn, self.loss_const1, shape=y_true.shape[0], y=y_true, **kwargs
        )
        loss_const2_value = _safe_run_lambda_array(
            self.loss_const2_fn, self.loss_const2, shape=y_true.shape[0], y=y_true, **kwargs
        )

        return gradient_const_value, loss_const1_value, loss_const2_value


class CostBoostGradientConst(SympyFnPickleMixin):
    """Class to compute the gradient constants of the cost metric for gradient boosting."""

    _sympy_functions: ClassVar[dict[str, str]] = {'gradient_const_fn': 'gradient_const_eq'}

    def __init__(self, tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr):
        y = sympy.symbols('y')
        self.gradient_const_eq = y * (-tp_benefit - fn_cost) + (1 - y) * (fp_cost + tn_benefit)
        self.gradient_const_fn = _safe_lambdify(self.gradient_const_eq)

    def __call__(self, y_true: FloatNDArray, **kwargs: Any) -> FloatNDArray:
        """Compute the gradient constants."""
        _check_parameters(self.gradient_const_eq.free_symbols - {sympy.symbols('y')}, kwargs)
        gradient_const_value = _safe_run_lambda_array(
            self.gradient_const_fn, self.gradient_const_eq, shape=y_true.shape[0], y=y_true, **kwargs
        )
        return gradient_const_value


class CostOptimalThreshold(SympyFnPickleMixin):
    """Class to compute the optimal threshold for the cost metric."""

    _sympy_functions: ClassVar[dict[str, str]] = {
        'calculate_denominator': 'denominator_expression',
        'calculate_numerator': 'numerator_expression',
    }

    def __init__(self, tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr):
        self.denominator_expression = fp_cost + tn_benefit + fn_cost + tp_benefit
        self.numerator_expression = fp_cost + tn_benefit
        self.calculate_denominator = _safe_lambdify(self.denominator_expression)
        self.calculate_numerator = _safe_lambdify(self.numerator_expression)

    def __call__(self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: Any) -> FloatNDArray | float:
        """Compute the optimal threshold(s). `y_true` and `y_score` are unused and kept for API compatibility."""
        _check_parameters(self.denominator_expression.free_symbols, parameters)
        denominator = _safe_run_lambda(self.calculate_denominator, self.denominator_expression, **parameters)
        numerator = _safe_run_lambda(self.calculate_numerator, self.numerator_expression, **parameters)

        eps = float(np.finfo(np.float64).eps)
        if np.isscalar(denominator):
            if denominator == 0:
                denominator = eps
        else:
            denominator = np.clip(denominator, float(np.finfo(float).eps), denominator)

        optimal = numerator / denominator
        return float(optimal) if np.isscalar(optimal) else optimal  # type: ignore[arg-type]


class CostOptimalRate(SympyFnPickleMixin):
    """Class to compute the optimal predicted positive rate for the cost metric."""

    _sympy_functions: ClassVar[dict[str, str]] = {
        'calculate_denominator': 'denominator_expression',
        'calculate_numerator': 'numerator_expression',
    }

    def __init__(self, tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr):
        self.denominator_expression = fp_cost + tn_benefit + fn_cost + tp_benefit
        self.numerator_expression = fp_cost + tn_benefit
        self.calculate_denominator = _safe_lambdify(self.denominator_expression)
        self.calculate_numerator = _safe_lambdify(self.numerator_expression)

    def __call__(self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: Any) -> float:
        """Compute the optimal predicted positive rate."""
        _check_parameters(self.denominator_expression.free_symbols, parameters)
        denominator = _safe_run_lambda(self.calculate_denominator, self.denominator_expression, **parameters)
        numerator = _safe_run_lambda(self.calculate_numerator, self.numerator_expression, **parameters)

        # Robust division to avoid divide-by-zero
        eps = float(np.finfo(np.float64).eps)
        if np.isscalar(denominator):
            denom_safe: FloatNDArray | float = denominator if denominator != 0 else eps  # type: ignore[assignment]
        else:
            denom_arr = np.asarray(denominator, dtype=np.float64)
            denom_safe = np.where(denom_arr == 0, eps, denom_arr)  # type: ignore[assignment]

        t_star = numerator / denom_safe

        # Normalize y\_score shape to 1D
        scores = np.asarray(y_score)
        if scores.ndim > 1:
            scores = scores.reshape(-1)

        # If t\* is scalar, compare against scalar; otherwise compare elementwise
        if np.isscalar(t_star):
            rate = float(np.mean(scores >= float(t_star)))  # type: ignore[arg-type]
        else:
            t_arr = np.asarray(t_star, dtype=np.float64).reshape(-1)
            rate = float(np.mean(scores >= t_arr))

        return rate


def _cost_loss_to_latex(
    tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr
) -> str:
    from sympy.printing.latex import latex

    i, N = sympy.symbols('i N')  # noqa: N806
    cost_function = (1 / N) * sympy.Sum(
        _format_cost_function(tp_cost=-tp_benefit, tn_cost=-tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost), (i, 0, N)
    )

    for symbol in cost_function.free_symbols:
        if symbol != N:
            cost_function = cost_function.subs(symbol, str(symbol) + '_i')

    output = latex(cost_function, mode='plain', order=None)

    return f'$\\displaystyle {output}$'


def _format_cost_function(
    tp_cost: sympy.Expr, tn_cost: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr
) -> sympy.Expr:
    y, s = sympy.symbols('y s')
    cost_function = y * (s * tp_cost + (1 - s) * fn_cost) + (1 - y) * ((1 - s) * tn_cost + s * fp_cost)
    return cost_function



from collections.abc import Callable
from numbers import Real

import numpy as np
import sympy

from ..._types import FloatArrayLike, FloatNDArray
from .common import Direction, _evaluate_expression
from .cost_matrix import CostMatrix
from .metric_strategies import MetricStrategy


class Metric:
    """
    Class to create a custom value/cost-sensitive metric.

    The metric is defined by a cost matrix and a strategy for computing the metric.
    The cost matrix defines the costs and benefits associated with each type of prediction outcome
    (true positive, true negative, false positive, false negative).
    The strategy defines how to compute the metric based on the cost matrix.

    Read more in the :ref:`User Guide <user_defined_value_metric>`.

    Parameters
    ----------
    cost_matrix : CostMatrix
        The cost matrix defining the costs and benefits associated with each type of prediction outcome.

    strategy : MetricStrategy
        The strategy to use for computing the metric.

        - If :class:`~empulse.metrics.MaxProfit`,
          the metric computes the maximum profit that can be achieved by a classifier.
          The metric determines the optimal threshold that maximizes the profit.
          This metric supports the use of stochastic variables.
        - If :class:`~empulse.metrics.Cost`, the metric computes the expected cost loss of a classifier.
          This metric supports passing instance-dependent costs in the form of array-likes.
          This metric does not support stochastic variables.
        - If :class:`~empulse.metrics.Savings`,
          the metric computes the savings that can be achieved by a classifier
          over a naive classifier which always predicts 0 or 1 (whichever is better).
          This metric supports passing instance-dependent costs in the form of array-likes.
          This metric does not support stochastic variables.

    Attributes
    ----------
    tp_benefit : sympy.Expr
        The benefit of a true positive.
        See :meth:`~empulse.metrics.Metric.add_tp_benefit` for more details.

    tn_benefit : sympy.Expr
        The benefit of a true negative.
        See :meth:`~empulse.metrics.Metric.add_tn_benefit` for more details.

    fp_benefit : sympy.Expr
        The benefit of a false positive.
        See :meth:`~empulse.metrics.Metric.add_fp_benefit` for more details.

    fn_benefit : sympy.Expr
        The benefit of a false negative.
        See :meth:`~empulse.metrics.Metric.add_fn_benefit` for more details.

    tp_cost : sympy.Expr
        The cost of a true positive.
        See :meth:`~empulse.metrics.Metric.add_tp_cost` for more details.

    tn_cost : sympy.Expr
        The cost of a true negative.
        See :meth:`~empulse.metrics.Metric.add_tn_cost` for more details.

    fp_cost : sympy.Expr
        The cost of a false positive.
        See :meth:`~empulse.metrics.Metric.add_fp_cost` for more details.

    fn_cost : sympy.Expr
        The cost of a false negative.
        See :meth:`~empulse.metrics.Metric.add_fn_cost` for more details.

    direction: Direction
        Whether the metric is to be maximized or minimized.

    Examples
    --------
    Reimplementing :func:`~empulse.metrics.empc_score` using the :class:`Metric` class.

    .. code-block:: python

        import sympy as sp
        from empulse.metrics import Metric, MaxProfit, CostMatrix

        clv, d, f, alpha, beta = sp.symbols(
            'clv d f alpha beta'
        )  # define deterministic variables
        gamma = sp.stats.Beta('gamma', alpha, beta)  # define gamma to follow a Beta distribution

        cost_matrix = (
            CostMatrix()
            .add_tp_benefit(gamma * (clv - d - f))  # when churner accepts offer
            .add_tp_benefit((1 - gamma) * -f)  # when churner does not accept offer
            .add_fp_cost(d + f)  # when you send an offer to a non-churner
            .alias({'incentive_cost': 'd', 'contact_cost': 'f'})
        )
        empc_score = Metric(cost_matrix, MaxProfit())

        y_true = [1, 0, 1, 0, 1]
        y_proba = [0.9, 0.1, 0.8, 0.2, 0.7]

        empc_score(y_true, y_proba, clv=100, incentive_cost=10, contact_cost=1, alpha=6, beta=14)

    Reimplementing :func:`~empulse.metrics.expected_cost_loss_churn` using the :class:`Metric` class.

    .. code-block:: python

        import sympy as sp
        from empulse.metrics import Metric, Cost, CostMatrix

        clv, delta, f, gamma = sp.symbols('clv delta f gamma')

        cost_matrix = (
            CostMatrix()
            .add_tp_benefit(gamma * (clv - delta * clv - f))  # when churner accepts offer
            .add_tp_benefit((1 - gamma) * -f)  # when churner does not accept offer
            .add_fp_cost(delta * clv + f)  # when you send an offer to a non-churner
            .alias({'incentive_fraction': 'delta', 'contact_cost': 'f', 'accept_rate': 'gamma'})
        )
        cost_loss = Metric(cost_matrix, Cost())

        y_true = [1, 0, 1, 0, 1]
        y_proba = [0.9, 0.1, 0.8, 0.2, 0.7]

        cost_loss(
            y_true, y_proba, clv=100, incentive_fraction=0.05, contact_cost=1, accept_rate=0.3
        )
    """

    def __init__(self, cost_matrix: CostMatrix, strategy: MetricStrategy) -> None:
        self.cost_matrix = cost_matrix
        self.strategy = strategy
        self.strategy.build(
            tp_benefit=self.tp_benefit,
            tn_benefit=self.tn_benefit,
            fp_cost=self.fp_cost,
            fn_cost=self.fn_cost,
        )

    @property
    def tp_benefit(self) -> sympy.Expr:  # noqa: D102
        return self.cost_matrix.tp_benefit

    @property
    def tn_benefit(self) -> sympy.Expr:  # noqa: D102
        return self.cost_matrix.tn_benefit

    @property
    def fp_benefit(self) -> sympy.Expr:  # noqa: D102
        return -self.cost_matrix.fp_cost

    @property
    def fn_benefit(self) -> sympy.Expr:  # noqa: D102
        return self.cost_matrix.fn_benefit

    @property
    def tp_cost(self) -> sympy.Expr:  # noqa: D102
        return self.cost_matrix.tp_cost

    @property
    def tn_cost(self) -> sympy.Expr:  # noqa: D102
        return self.cost_matrix.tn_cost

    @property
    def fp_cost(self) -> sympy.Expr:  # noqa: D102
        return self.cost_matrix.fp_cost

    @property
    def fn_cost(self) -> sympy.Expr:  # noqa: D102
        return self.cost_matrix.fn_cost

    @property
    def direction(self) -> Direction:  # noqa: D102
        return self.strategy.direction

    @property
    def _all_symbols(self) -> set[str]:
        """Return a set of all symbols used in the cost matrix."""
        all_symbols = (
            self.tp_cost.free_symbols
            | self.tn_cost.free_symbols
            | self.fp_cost.free_symbols
            | self.fn_cost.free_symbols
            | self.cost_matrix._aliases.keys()
        )
        return {str(symbol) for symbol in all_symbols}

    def _prepare_parameters(self, **kwargs: FloatArrayLike | float) -> dict[str, FloatNDArray | float]:
        """Swap aliases with the appropriate symbols and convert the values to numpy arrays."""
        # Map aliases to the appropriate symbols
        for alias, symbol in self.cost_matrix._aliases.items():
            if alias in kwargs:
                kwargs[symbol] = kwargs.pop(alias)

        # Use default values if not provided in kwargs
        for key, value in self.cost_matrix._defaults.items():
            kwargs.setdefault(key, value)

        for key, value in kwargs.items():
            if not isinstance(value, Real | str):
                kwargs[key] = np.asarray(value).reshape(-1)

        # convert any ints to floats
        for key, value in kwargs.items():
            if isinstance(value, np.ndarray) and not np.issubdtype(value.dtype, np.floating):
                kwargs[key] = value.astype(np.float64)
            elif isinstance(value, int):
                kwargs[key] = float(value)

        kwargs: dict[str, FloatNDArray | float]  # redefine kwargs as mypy doesn't understand the above

        return kwargs

    def __call__(self, y_true: FloatArrayLike, y_score: FloatArrayLike, **parameters: FloatArrayLike | float) -> float:
        """
        Compute the metric score or loss.

        The :meth:`empulse.metrics.Metric.build` method should be called before calling this method.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

            - If :class:`~empulse.metrics.MaxProfit`, the predicted labels are the decision scores.
            - If :class:`~empulse.metrics.Cost`, the predicted labels are the (calibrated) probabilities.
            - If :class:`~empulse.metrics.Savings`, the predicted labels are the (calibrated) probabilities.

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        score: float
            The computed metric score or loss.
        """
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
        parameters = self._prepare_parameters(**parameters)
        return self.strategy.score(y_true, y_score, **parameters)

    def optimal_threshold(
        self, y_true: FloatArrayLike, y_score: FloatArrayLike, **parameters: FloatArrayLike | float
    ) -> FloatNDArray | float:
        """
        Compute the optimal classification threshold(s).

        i.e., the score threshold at which an observation should be classified as positive to optimize the metric.
        For instance-dependent costs and benefits, this will return an array of thresholds, one for each sample.
        For class-dependent costs and benefits, this will return a single threshold value.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

            - If :class:`~empulse.metrics.MaxProfit`, the predicted labels are the decision scores.
            - If :class:`~empulse.metrics.Cost`, the predicted labels are the (calibrated) probabilities.
            - If :class:`~empulse.metrics.Savings`, the predicted labels are the (calibrated) probabilities.

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        optimal_threshold: float or NDArray of shape (n_samples,)
            The optimal classification threshold(s).
        """
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
        parameters = self._prepare_parameters(**parameters)
        return self.strategy.optimal_threshold(y_true, y_score, **parameters)

    def optimal_rate(
        self, y_true: FloatArrayLike, y_score: FloatArrayLike, **parameters: FloatArrayLike | float
    ) -> float:
        """
        Compute the optimal predicted positive rate.

        i.e., the fraction of observations that should be classified as positive to optimize the metric.

        Parameters
        ----------
        y_true: array-like of shape (n_samples,)
            The ground truth labels.

        y_score: array-like of shape (n_samples,)
            The predicted labels, probabilities, or decision scores (based on the chosen metric).

            - If :class:`~empulse.metrics.MaxProfit`, the predicted labels are the decision scores.
            - If :class:`~empulse.metrics.Cost`, the predicted labels are the (calibrated) probabilities.
            - If :class:`~empulse.metrics.Savings`, the predicted labels are the (calibrated) probabilities.

        parameters: float or array-like of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        optimal_rate: float
            The optimal predicted positive rate.
        """
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
        parameters = self._prepare_parameters(**parameters)
        return self.strategy.optimal_rate(y_true, y_score, **parameters)

    def _logit_objective(
        self,
        features: FloatNDArray,
        y_true: FloatNDArray,
        C: float,
        l1_ratio: float,
        soft_threshold: bool,
        fit_intercept: bool,
        **parameters: FloatNDArray | float,
    ) -> Callable[[FloatNDArray], tuple[float, FloatNDArray]]:
        """
        Compute the metric loss and its gradient with respect to the logistic regression weights.

        Parameters
        ----------
        features : NDArray of shape (n_samples, n_features)
            The features of the samples.
        weights : NDArray of shape (n_features,)
            The weights of the logistic regression model.
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        value : float
            The metric loss to be minimized.
        gradient : NDArray of shape (n_features,)
            The gradient of the metric loss with respect to the logistic regression weights.
        """
        parameters = self._prepare_parameters(**parameters)

        if y_true.ndim == 1:
            y_true = np.expand_dims(y_true, axis=1)
        for key, value in parameters.items():
            if isinstance(value, np.ndarray) and value.ndim == 1:
                parameters[key] = np.expand_dims(value, axis=1)

        return self.strategy.logit_objective(
            features=features,
            y_true=y_true,
            C=C,
            l1_ratio=l1_ratio,
            soft_threshold=soft_threshold,
            fit_intercept=fit_intercept,
            **parameters,
        )

    def _prepare_logit_objective(
        self, features: FloatNDArray, y_true: FloatNDArray, **parameters: FloatNDArray | float
    ) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray]:
        """
        Compute the constant term of the loss and gradient of the metric wrt logistic regression coefficients.

        Parameters
        ----------
        features : NDArray of shape (n_samples, n_features)
            The features of the samples.
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        gradient_const : NDArray of shape (n_samples, n_features)
            The constant term of the gradient.
        loss_const1 : NDArray of shape (n_features,)
            The first constant term of the loss function.
        loss_const2 : NDArray of shape (n_features,)
            The second constant term of the loss function.
        """
        parameters = self._prepare_parameters(**parameters)
        for key, value in parameters.items():
            if isinstance(value, np.ndarray) and value.ndim == 1:
                parameters[key] = np.expand_dims(value, axis=1)
        return self.strategy.prepare_logit_objective(features, y_true, **parameters)

    def _gradient_boost_objective(
        self, y_true: FloatNDArray, y_score: FloatNDArray, **parameters: FloatNDArray | float
    ) -> tuple[FloatNDArray, FloatNDArray]:
        """
        Compute the gradient and hessian of the metric loss with respect to the gradient boosting weights.

        Parameters
        ----------
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        y_score : NDArray of shape (n_samples,)
            The predicted probabilities or decision scores.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        gradient : NDArray of shape (n_samples,)
            The gradient of the metric loss with respect to the gradient boosting weights.
        hessian : NDArray of shape (n_samples,)
            The hessian of the metric loss with respect to the gradient boosting weights.
        """
        parameters = self._prepare_parameters(**parameters)
        # y_proba = scipy.special.expit(y_score)
        y_proba = y_score
        gradient, hessian = self.strategy.gradient_boost_objective(y_true, y_proba, **parameters)
        return gradient, hessian

    def _prepare_boost_objective(self, y_true: FloatNDArray, **parameters: FloatNDArray | float) -> FloatNDArray:
        """
        Compute the gradient's constant term of the metric wrt gradient boost.

        Parameters
        ----------
        y_true : NDArray of shape (n_samples,)
            The ground truth labels.
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

            - If ``float``, the same value is used for all samples (class-dependent).
            - If ``array-like``, the values are used for each sample (instance-dependent).

        Returns
        -------
        gradient_const : NDArray of shape (n_samples, n_features)
            The constant term of the gradient.
        """
        parameters = self._prepare_parameters(**parameters)
        for key, value in parameters.items():
            if isinstance(value, np.ndarray) and value.ndim == 1:
                parameters[key] = np.expand_dims(value, axis=1)
        return self.strategy.prepare_boost_objective(y_true, **parameters)

    def _evaluate_costs(
        self, **parameters: FloatNDArray | float
    ) -> tuple[
        FloatNDArray | float,
        FloatNDArray | float,
        FloatNDArray | float,
        FloatNDArray | float,
    ]:
        """
        Evaluate the costs expressions.

        Parameters
        ----------
        parameters : float or NDArray of shape (n_samples,)
            The parameter values for the costs and benefits defined in the metric.
            If any parameter is a stochastic variable, you should pass values for their distribution parameters.
            You can set the parameter values for either the symbol names or their aliases.

        Returns
        -------
        fp_cost : float or NDArray of shape (n_samples,)
            The false positive cost(s).
        fn_cost : float or NDArray of shape (n_samples,)
            The false negative cost(s).
        tp_cost : float or NDArray of shape (n_samples,)
            The true positive cost(s).
        tn_cost : float or NDArray of shape (n_samples,)
            The true negative cost(s).
        """
        parameters = self._prepare_parameters(**parameters)
        fp_cost = _evaluate_expression(self.fp_cost, **parameters)
        fn_cost = _evaluate_expression(self.fn_cost, **parameters)
        tp_cost = _evaluate_expression(self.tp_cost, **parameters)
        tn_cost = _evaluate_expression(self.tn_cost, **parameters)
        return fp_cost, fn_cost, tp_cost, tn_cost

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(cost_matrix={self.cost_matrix}, strategy={self.strategy})'

    def _repr_latex_(self) -> str:
        return self.strategy.to_latex(
            tp_benefit=self.tp_benefit, tn_benefit=self.tn_benefit, fp_cost=self.fp_cost, fn_cost=self.fn_cost
        )



from functools import lru_cache

import numpy as np

from ..._types import FloatArrayLike, FloatNDArray
from ..common import _compute_profits
from ._validation import _validate_input_deterministic


def mpa_score(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    contribution: float = 8_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
    check_input: bool = True,
) -> float:
    """
    :func:`~empulse.metrics.mpa()` but only returning the MPA score.

    MPA presumes a situation where leads are targeted either directly or indirectly.
    Directly targeted leads are contacted and handled by the internal sales team.
    Indirectly targeted leads are contacted and then referred to intermediaries,
    which receive a commission.
    The company gains a contribution from a successful acquisition.

    .. seealso::

        :func:`~empulse.metrics.mpa` : to also return the fraction of the leads
        that should be targeted to maximize profit.

        :func:`~empulse.metrics.empa_score` : for a stochastic version of this metric.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        Binary target values ('acquisition': 1, 'no acquisition': 0).

    y_score : array-like of shape (n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    contribution : float, default=7000
        Average contribution of a new customer (``contribution ≥ 0``).

    sales_cost : float, default=500
        Average sale conversion cost of targeted leads handled by the company (``sales_cost ≥ 0``).

    contact_cost : float, default=50
        Average contact cost of targeted leads (``contact_cost ≥ 0``).

    direct_selling : float, default=1
        Fraction of leads sold to directly (``0 ≤ direct_selling ≤ 1``).
        ``direct_selling = 0`` for indirect channel.
        ``direct_selling = 1`` for direct channel.

    commission : float, default=0.1
        Fraction of contribution paid to the intermediaries (``0 ≤ commission ≤ 1``).

        .. note::
            The commission is only relevant when there is an indirect channel (``direct_selling < 1``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empa : float
        Expected Maximum Profit measure for customer Acquisition.

    Examples
    --------
    Direct channel:

    >>> from empulse.metrics import mpa_score
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> mpa_score(y_true, y_score, direct_selling=1)
    3706.25

    Indirect channel using scorer:

    >>> import numpy as np
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import cross_val_score, StratifiedKFold
    >>> from sklearn.metrics import make_scorer
    >>> from empulse.metrics import mpa_score
    >>>
    >>> X, y = make_classification(random_state=42)
    >>> model = LogisticRegression()
    >>> cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)
    >>> scorer = make_scorer(
    ...     mpa_score,
    ...     response_method='predict_proba',
    ...     contribution=7_000,
    ...     sales_cost=2_000,
    ...     contact_cost=100,
    ...     direct_selling=0,
    ... )
    >>> np.mean(cross_val_score(model, X, y, cv=cv, scoring=scorer))
    3099.0
    """
    return mpa(
        y_true,
        y_score,
        contribution=contribution,
        contact_cost=contact_cost,
        sales_cost=sales_cost,
        direct_selling=direct_selling,
        commission=commission,
        check_input=check_input,
    )[0]


def mpa(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    contribution: float = 8_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
    check_input: bool = True,
) -> tuple[float, float]:
    """
    Maximum Profit measure for customer Acquisition (MPA).

    MPA presumes a situation where leads are targeted either directly or indirectly.
    Directly targeted leads are contacted and handled by the internal sales team.
    Indirectly targeted leads are contacted and then referred to intermediaries,
    which receive a commission.
    The company gains a contribution from a successful acquisition.

    .. seealso::

        :func:`~empulse.metrics.mpa_score` : to only return the EMPA score.

        :func:`~empulse.metrics.empa` : for a stochastic version of this metric.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        Binary target values ('acquisition': 1, 'no acquisition': 0).

    y_score : array-like of shape (n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    contribution : float, default=7000
        Average contribution of a new customer (``contribution ≥ 0``).

    sales_cost : float, default=500
        Average sale conversion cost of targeted leads handled by the company (``sales_cost ≥ 0``).

    contact_cost : float, default=50
        Average contact cost of targeted leads (``contact_cost ≥ 0``).

    direct_selling : float, default=1
        Fraction of leads sold to directly (``0 ≤ direct_selling ≤ 1``).
        ``direct_selling = 0`` for indirect channel.
        ``direct_selling = 1`` for direct channel.

    commission : float, default=0.1
        Fraction of contribution paid to the intermediaries (``0 ≤ commission ≤ 1``).

        .. note::
            The commission is only relevant when there is an indirect channel (``direct_selling < 1``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    mpa : float
        Maximum Profit measure for customer Acquisition

    threshold : float
        Fraction of the leads that should be targeted to maximize profit

    Examples
    --------
    >>> from empulse.metrics import mpa
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> mpa(y_true, y_score)
    (3706.25, 0.875)
    """
    profits, customer_thresholds = _compute_profit_acquisition(
        y_true,
        y_score,
        contribution,
        contact_cost,
        sales_cost,
        direct_selling,
        commission,
        check_input,
    )
    max_profit_index = np.argmax(profits)
    return profits[max_profit_index], customer_thresholds[max_profit_index]


def _compute_profit_acquisition(
    y_true: FloatArrayLike,
    y_pred: FloatArrayLike,
    contribution: float = 8_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
    check_input: bool = True,
) -> tuple[FloatNDArray, FloatNDArray]:
    if check_input:
        y_true, y_pred = _validate_input_deterministic(
            y_true, y_pred, contribution, contact_cost, sales_cost, direct_selling, commission
        )
    else:
        y_true = np.asarray(y_true)
        y_pred = np.asarray(y_pred)
    cost_benefits = _compute_cost_benefits(contribution, contact_cost, sales_cost, direct_selling, commission)
    return _compute_profits(y_true, y_pred, cost_benefits)


@lru_cache(maxsize=1)
def _compute_cost_benefits(
    contribution: float,
    contact_cost: float,
    sales_cost: float,
    direct_selling: float,
    commission: float,
) -> FloatNDArray:
    true_positive_benefit = direct_selling * (contribution - contact_cost - sales_cost) + (1 - direct_selling) * (
        (1 - commission) * contribution - contact_cost
    )
    false_positive_cost = -contact_cost
    return np.array([true_positive_benefit, false_positive_cost])



cimport numpy as cnp
import numpy as np
from sklearn.utils._typedefs cimport float64_t, intp_t
from sklearn.tree._criterion cimport ClassificationCriterion
from libc.math cimport fmin
from libc.math cimport log as ln
from libc.string cimport memcpy
from libc.string cimport memset


cdef inline float64_t log(float64_t x) noexcept nogil:
    return ln(x) / ln(2.0)


cdef class CostImpurity(ClassificationCriterion):
    cdef float64_t tp_cost, tn_cost, fp_cost, fn_cost
    cdef float64_t[:] tp_cost_array, tn_cost_array, fp_cost_array, fn_cost_array
    cdef bint all_class_dependent

    # New: cost accumulators
    cdef float64_t[:, ::1] pos_cost_sum_total
    cdef float64_t[:, ::1] neg_cost_sum_total
    cdef float64_t[:, ::1] pos_cost_sum_left
    cdef float64_t[:, ::1] neg_cost_sum_left
    cdef float64_t[:, ::1] pos_cost_sum_right
    cdef float64_t[:, ::1] neg_cost_sum_right

    def __deepcopy__(self, memo):
        # Convert memoryview to numpy array
        n_classes_array = np.asarray(self.n_classes)
        new_instance = CostImpurity(self.n_outputs, n_classes_array)

        # Copy all attributes (otherwise not included)
        new_instance.tp_cost = self.tp_cost
        new_instance.tn_cost = self.tn_cost
        new_instance.fp_cost = self.fp_cost
        new_instance.fn_cost = self.fn_cost
        new_instance.tp_cost_array = np.array(self.tp_cost_array)
        new_instance.tn_cost_array = np.array(self.tn_cost_array)
        new_instance.fp_cost_array = np.array(self.fp_cost_array)
        new_instance.fn_cost_array = np.array(self.fn_cost_array)
        new_instance.all_class_dependent = self.all_class_dependent
        new_instance.pos_cost_sum_total = np.array(self.pos_cost_sum_total)
        new_instance.neg_cost_sum_total = np.array(self.neg_cost_sum_total)
        new_instance.pos_cost_sum_left = np.array(self.pos_cost_sum_left)
        new_instance.neg_cost_sum_left = np.array(self.neg_cost_sum_left)
        new_instance.pos_cost_sum_right = np.array(self.pos_cost_sum_right)
        new_instance.neg_cost_sum_right = np.array(self.neg_cost_sum_right)

        # Copy other attributes if necessary
        memo[id(self)] = new_instance
        return new_instance

    def set_costs(self, float64_t tp_cost=0.0, float64_t tn_cost=0.0, float64_t fp_cost=0.0, float64_t fn_cost=0.0):
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fp_cost = fp_cost
        self.fn_cost = fn_cost

    def set_array_costs(self, float64_t[:] tp_cost, float64_t[:] tn_cost, float64_t[:] fp_cost, float64_t[:] fn_cost, intp_t n_samples):
        cdef bint is_tp_cost_array, is_tn_cost_array, is_fp_cost_array, is_fn_cost_array
        is_tp_cost_array = tp_cost.shape[0] > 0
        is_tn_cost_array = tn_cost.shape[0] > 0
        is_fp_cost_array = fp_cost.shape[0] > 0
        is_fn_cost_array = fn_cost.shape[0] > 0
        self.all_class_dependent = not (
            is_tp_cost_array or
            is_tn_cost_array or
            is_fp_cost_array or
            is_fn_cost_array
        )
        if is_tp_cost_array:
            self.tp_cost_array = tp_cost
        else:
            self.tp_cost_array = np.full(n_samples, self.tp_cost, dtype=np.float64)
        if is_tn_cost_array:
            self.tn_cost_array = tn_cost
        else:
            self.tn_cost_array = np.full(n_samples, self.tn_cost, dtype=np.float64)
        if is_fp_cost_array:
            self.fp_cost_array = fp_cost
        else:
            self.fp_cost_array = np.full(n_samples, self.fp_cost, dtype=np.float64)
        if is_fn_cost_array:
            self.fn_cost_array = fn_cost
        else:
            self.fn_cost_array = np.full(n_samples, self.fn_cost, dtype=np.float64)
        self.pos_cost_sum_total = np.zeros((self.n_outputs, self.max_n_classes), dtype=np.float64)
        self.neg_cost_sum_total = np.zeros((self.n_outputs, self.max_n_classes), dtype=np.float64)
        self.pos_cost_sum_left = np.zeros((self.n_outputs, self.max_n_classes), dtype=np.float64)
        self.neg_cost_sum_left = np.zeros((self.n_outputs, self.max_n_classes), dtype=np.float64)
        self.pos_cost_sum_right = np.zeros((self.n_outputs, self.max_n_classes), dtype=np.float64)
        self.neg_cost_sum_right = np.zeros((self.n_outputs, self.max_n_classes), dtype=np.float64)

    cdef int init(
        self,
        const float64_t[:, ::1] y,
        const float64_t[:] sample_weight,
        float64_t weighted_n_samples,
        const intp_t[:] sample_indices,
        intp_t start,
        intp_t end,
    ) except -1 nogil:
        # Standard class counts
        ClassificationCriterion.init(self, y, sample_weight, weighted_n_samples, sample_indices, start, end)
        # Cost accumulators
        cdef intp_t k, c, i, p
        cdef float64_t w = 1.0
        memset(&self.pos_cost_sum_total[0, 0], 0, self.n_outputs * self.max_n_classes * sizeof(float64_t))
        memset(&self.neg_cost_sum_total[0, 0], 0, self.n_outputs * self.max_n_classes * sizeof(float64_t))
        for p in range(start, end):
            i = sample_indices[p]

            for k in range(self.n_outputs):
                c = <intp_t> y[i, k]
                if self.all_class_dependent:
                    if c == 1:
                        self.pos_cost_sum_total[k, c] += self.tp_cost
                        self.neg_cost_sum_total[k, c] += self.fn_cost
                    else:
                        self.pos_cost_sum_total[k, c] += self.fp_cost
                        self.neg_cost_sum_total[k, c] += self.tn_cost
                else:
                    if c == 1:
                        self.pos_cost_sum_total[k, c] += self.tp_cost_array[i]
                        self.neg_cost_sum_total[k, c] += self.fn_cost_array[i]
                    else:
                        self.pos_cost_sum_total[k, c] += self.fp_cost_array[i]
                        self.neg_cost_sum_total[k, c] += self.tn_cost_array[i]
        # Reset left/right
        self.reset()
        return 0

    cdef int reset(self) except -1 nogil:
        self.pos = self.start
        memset(&self.pos_cost_sum_left[0, 0], 0, self.n_outputs * self.max_n_classes * sizeof(float64_t))
        memset(&self.neg_cost_sum_left[0, 0], 0, self.n_outputs * self.max_n_classes * sizeof(float64_t))
        memcpy(&self.pos_cost_sum_right[0, 0], &self.pos_cost_sum_total[0, 0], self.n_outputs * self.max_n_classes * sizeof(float64_t))
        memcpy(&self.neg_cost_sum_right[0, 0], &self.neg_cost_sum_total[0, 0], self.n_outputs * self.max_n_classes * sizeof(float64_t))
        ClassificationCriterion.reset(self)
        return 0

    cdef int reverse_reset(self) except -1 nogil:
        self.pos = self.end
        memset(&self.pos_cost_sum_right[0, 0], 0, self.n_outputs * self.max_n_classes * sizeof(float64_t))
        memset(&self.neg_cost_sum_right[0, 0], 0, self.n_outputs * self.max_n_classes * sizeof(float64_t))
        memcpy(&self.pos_cost_sum_left[0, 0], &self.pos_cost_sum_total[0, 0], self.n_outputs * self.max_n_classes * sizeof(float64_t))
        memcpy(&self.neg_cost_sum_left[0, 0], &self.neg_cost_sum_total[0, 0], self.n_outputs * self.max_n_classes * sizeof(float64_t))
        ClassificationCriterion.reverse_reset(self)
        return 0

    cdef int update(self, intp_t new_pos) except -1 nogil:
        cdef intp_t pos = self.pos
        cdef intp_t end_non_missing = self.end - self.n_missing
        cdef const intp_t[:] sample_indices = self.sample_indices
        cdef const float64_t[:] sample_weight = self.sample_weight
        cdef intp_t i, p, k, c
        cdef float64_t w = 1.0
        if (new_pos - pos) <= (end_non_missing - new_pos):
            for p in range(pos, new_pos):
                i = sample_indices[p]
                if sample_weight is not None:
                    w = sample_weight[i]
                for k in range(self.n_outputs):
                    c = <intp_t> self.y[i, k]
                    self.sum_left[k, c] += w
                    if self.all_class_dependent:
                        if c == 1:
                            self.pos_cost_sum_left[k, c] += w * self.tp_cost
                            self.neg_cost_sum_left[k, c] += w * self.fn_cost
                        else:
                            self.pos_cost_sum_left[k, c] += w * self.fp_cost
                            self.neg_cost_sum_left[k, c] += w * self.tn_cost
                    else:
                        if c == 1:
                            self.pos_cost_sum_left[k, c] += w * self.tp_cost_array[i]
                            self.neg_cost_sum_left[k, c] += w * self.fn_cost_array[i]
                        else:
                            self.pos_cost_sum_left[k, c] += w * self.fp_cost_array[i]
                            self.neg_cost_sum_left[k, c] += w * self.tn_cost_array[i]
                self.weighted_n_left += w
        else:
            self.reverse_reset()
            for p in range(end_non_missing - 1, new_pos - 1, -1):
                i = sample_indices[p]
                if sample_weight is not None:
                    w = sample_weight[i]
                for k in range(self.n_outputs):
                    c = <intp_t> self.y[i, k]
                    self.sum_left[k, c] -= w
                    if self.all_class_dependent:
                        if c == 1:
                            self.pos_cost_sum_left[k, c] -= w * self.tp_cost
                            self.neg_cost_sum_left[k, c] -= w * self.fn_cost
                        else:
                            self.pos_cost_sum_left[k, c] -= w * self.fp_cost
                            self.neg_cost_sum_left[k, c] -= w * self.tn_cost
                    else:
                        if c == 1:
                            self.pos_cost_sum_left[k, c] -= w * self.tp_cost_array[i]
                            self.neg_cost_sum_left[k, c] -= w * self.fn_cost_array[i]
                        else:
                            self.pos_cost_sum_left[k, c] -= w * self.fp_cost_array[i]
                            self.neg_cost_sum_left[k, c] -= w * self.tn_cost_array[i]
                self.weighted_n_left -= w
        # Update right
        self.weighted_n_right = self.weighted_n_node_samples - self.weighted_n_left
        for k in range(self.n_outputs):
            for c in range(self.n_classes[k]):
                self.sum_right[k, c] = self.sum_total[k, c] - self.sum_left[k, c]
                self.pos_cost_sum_right[k, c] = self.pos_cost_sum_total[k, c] - self.pos_cost_sum_left[k, c]
                self.neg_cost_sum_right[k, c] = self.neg_cost_sum_total[k, c] - self.neg_cost_sum_left[k, c]
        self.pos = new_pos
        return 0

    cdef float64_t node_impurity(self) noexcept nogil:
        cdef float64_t pos_cost = 0.0
        cdef float64_t neg_cost = 0.0
        cdef intp_t k, c
        for k in range(self.n_outputs):
            for c in range(self.n_classes[k]):
                pos_cost += self.pos_cost_sum_total[k, c]
                neg_cost += self.neg_cost_sum_total[k, c]
        return fmin(pos_cost, neg_cost) / self.weighted_n_node_samples

    cdef void children_impurity(self, float64_t* impurity_left, float64_t* impurity_right) noexcept nogil:
        cdef float64_t pos_cost_left = 0.0
        cdef float64_t neg_cost_left = 0.0
        cdef float64_t pos_cost_right = 0.0
        cdef float64_t neg_cost_right = 0.0
        cdef intp_t k, c
        for k in range(self.n_outputs):
            for c in range(self.n_classes[k]):
                pos_cost_left += self.pos_cost_sum_left[k, c]
                neg_cost_left += self.neg_cost_sum_left[k, c]
                pos_cost_right += self.pos_cost_sum_right[k, c]
                neg_cost_right += self.neg_cost_sum_right[k, c]
        impurity_left[0] = fmin(pos_cost_left, neg_cost_left) / self.weighted_n_left
        impurity_right[0] = fmin(pos_cost_right, neg_cost_right) / self.weighted_n_right


cdef class GiniCostImpurity(CostImpurity):

    def __deepcopy__(self, memo):
        # Convert memoryview to numpy array
        n_classes_array = np.asarray(self.n_classes)
        new_instance = GiniCostImpurity(self.n_outputs, n_classes_array)

        # Copy all attributes (otherwise not included)
        new_instance.tp_cost = self.tp_cost
        new_instance.tn_cost = self.tn_cost
        new_instance.fp_cost = self.fp_cost
        new_instance.fn_cost = self.fn_cost
        new_instance.tp_cost_array = np.array(self.tp_cost_array)
        new_instance.tn_cost_array = np.array(self.tn_cost_array)
        new_instance.fp_cost_array = np.array(self.fp_cost_array)
        new_instance.fn_cost_array = np.array(self.fn_cost_array)
        new_instance.all_class_dependent = self.all_class_dependent
        new_instance.pos_cost_sum_total = np.array(self.pos_cost_sum_total)
        new_instance.neg_cost_sum_total = np.array(self.neg_cost_sum_total)
        new_instance.pos_cost_sum_left = np.array(self.pos_cost_sum_left)
        new_instance.neg_cost_sum_left = np.array(self.neg_cost_sum_left)
        new_instance.pos_cost_sum_right = np.array(self.pos_cost_sum_right)
        new_instance.neg_cost_sum_right = np.array(self.neg_cost_sum_right)

        # Copy other attributes if necessary
        memo[id(self)] = new_instance
        return new_instance

    cdef float64_t node_impurity(self) noexcept nogil:
        cdef float64_t pos_sq_prior, neg_sq_prior
        cdef float64_t pos_count, neg_count
        cdef float64_t pos_cost = 0.0
        cdef float64_t neg_cost = 0.0
        cdef intp_t k, c
        for k in range(self.n_outputs):
            for c in range(self.n_classes[k]):
                pos_cost += self.pos_cost_sum_total[k, c]
                neg_cost += self.neg_cost_sum_total[k, c]
        pos_count = self.sum_total[0, 1]
        pos_sq_prior = pos_count * pos_count / self.weighted_n_node_samples
        neg_count = self.sum_total[0, 0]
        neg_sq_prior = neg_count * neg_count / self.weighted_n_node_samples
        return fmin(pos_cost * pos_sq_prior, neg_cost * neg_sq_prior) / self.weighted_n_node_samples

    cdef void children_impurity(self, float64_t* impurity_left, float64_t* impurity_right) noexcept nogil:
        cdef float64_t pos_sq_prior_left, neg_sq_prior_left, pos_sq_prior_right, neg_sq_prior_right
        cdef float64_t pos_count_left, neg_count_left, pos_count_right, neg_count_right
        cdef float64_t pos_cost_left = 0.0
        cdef float64_t neg_cost_left = 0.0
        cdef float64_t pos_cost_right = 0.0
        cdef float64_t neg_cost_right = 0.0
        cdef intp_t k, c
        for k in range(self.n_outputs):
            for c in range(self.n_classes[k]):
                pos_cost_left += self.pos_cost_sum_left[k, c]
                neg_cost_left += self.neg_cost_sum_left[k, c]
                pos_cost_right += self.pos_cost_sum_right[k, c]
                neg_cost_right += self.neg_cost_sum_right[k, c]

        pos_count_left = self.sum_left[0, 1]
        pos_sq_prior_left = pos_count_left * pos_count_left / self.weighted_n_left
        neg_count_left = self.sum_left[0, 0]
        neg_sq_prior_left = neg_count_left * neg_count_left / self.weighted_n_left

        pos_count_right = self.sum_right[0, 1]
        pos_sq_prior_right = pos_count_right * pos_count_right / self.weighted_n_right
        neg_count_right = self.sum_right[0, 0]
        neg_sq_prior_right = neg_count_right * neg_count_right / self.weighted_n_right

        impurity_left[0] = fmin(pos_cost_left * pos_sq_prior_left, neg_cost_left * neg_sq_prior_left) / self.weighted_n_left
        impurity_right[0] = fmin(pos_cost_right * pos_sq_prior_right, neg_cost_right * neg_sq_prior_right) / self.weighted_n_right


cdef class EntropyCostImpurity(CostImpurity):

    def __deepcopy__(self, memo):
        # Convert memoryview to numpy array
        n_classes_array = np.asarray(self.n_classes)
        new_instance = EntropyCostImpurity(self.n_outputs, n_classes_array)

        # Copy all attributes (otherwise not included)
        new_instance.tp_cost = self.tp_cost
        new_instance.tn_cost = self.tn_cost
        new_instance.fp_cost = self.fp_cost
        new_instance.fn_cost = self.fn_cost
        new_instance.tp_cost_array = np.array(self.tp_cost_array)
        new_instance.tn_cost_array = np.array(self.tn_cost_array)
        new_instance.fp_cost_array = np.array(self.fp_cost_array)
        new_instance.fn_cost_array = np.array(self.fn_cost_array)
        new_instance.all_class_dependent = self.all_class_dependent
        new_instance.pos_cost_sum_total = np.array(self.pos_cost_sum_total)
        new_instance.neg_cost_sum_total = np.array(self.neg_cost_sum_total)
        new_instance.pos_cost_sum_left = np.array(self.pos_cost_sum_left)
        new_instance.neg_cost_sum_left = np.array(self.neg_cost_sum_left)
        new_instance.pos_cost_sum_right = np.array(self.pos_cost_sum_right)
        new_instance.neg_cost_sum_right = np.array(self.neg_cost_sum_right)

        # Copy other attributes if necessary
        memo[id(self)] = new_instance
        return new_instance

    cdef float64_t node_impurity(self) noexcept nogil:
        cdef float64_t pos_entropy = 0.0
        cdef float64_t neg_entropy = 0.0
        cdef float64_t pos_count, neg_count
        cdef float64_t pos_cost = 0.0
        cdef float64_t neg_cost = 0.0
        cdef intp_t k, c
        for k in range(self.n_outputs):
            for c in range(self.n_classes[k]):
                pos_cost += self.pos_cost_sum_total[k, c]
                neg_cost += self.neg_cost_sum_total[k, c]
        pos_cost /= self.weighted_n_node_samples
        neg_cost /= self.weighted_n_node_samples

        pos_count = self.sum_total[0, 1]
        if pos_count > 0.0:
            pos_entropy = log(pos_count / self.weighted_n_node_samples)
        neg_count = self.sum_total[0, 0]
        if neg_count > 0.0:
            neg_entropy = log(neg_count / self.weighted_n_node_samples)
        return fmin(pos_cost * -pos_entropy, neg_cost * -neg_entropy)

    cdef void children_impurity(self, float64_t* impurity_left, float64_t* impurity_right) noexcept nogil:
        cdef float64_t pos_entropy_left = 0.0
        cdef float64_t neg_entropy_left = 0.0
        cdef float64_t pos_entropy_right = 0.0
        cdef float64_t neg_entropy_right = 0.0
        cdef float64_t pos_count_left, neg_count_left, pos_count_right, neg_count_right
        cdef float64_t pos_cost_left = 0.0
        cdef float64_t neg_cost_left = 0.0
        cdef float64_t pos_cost_right = 0.0
        cdef float64_t neg_cost_right = 0.0
        cdef intp_t k, c
        for k in range(self.n_outputs):
            for c in range(self.n_classes[k]):
                pos_cost_left += self.pos_cost_sum_left[k, c]
                neg_cost_left += self.neg_cost_sum_left[k, c]
                pos_cost_right += self.pos_cost_sum_right[k, c]
                neg_cost_right += self.neg_cost_sum_right[k, c]
        pos_cost_left /= self.weighted_n_left
        pos_cost_right /= self.weighted_n_right
        neg_cost_left /= self.weighted_n_left
        neg_cost_right /= self.weighted_n_right

        pos_count_left = self.sum_left[0, 1]
        if pos_count_left > 0.0:
            pos_entropy_left = log(pos_count_left / self.weighted_n_left)
        neg_count_left = self.sum_left[0, 0]
        if neg_count_left > 0.0:
            neg_entropy_left = log(neg_count_left / self.weighted_n_left)
        pos_count_right = self.sum_right[0, 1]
        if pos_count_right > 0.0:
            pos_entropy_right = log(pos_count_right / self.weighted_n_right)
        neg_count_right = self.sum_right[0, 0]
        if neg_count_right > 0.0:
            neg_entropy_right = log(neg_count_right / self.weighted_n_right)

        impurity_left[0] = fmin(pos_cost_left * pos_entropy_left, neg_cost_left * neg_entropy_left)
        impurity_right[0] = fmin(pos_cost_right * pos_entropy_right, neg_cost_right * neg_entropy_right)


import sys
from collections.abc import MutableMapping
from numbers import Real
from typing import Any, ClassVar, Literal, TypeVar

import numpy as np
import scipy.stats as st
from numpy.typing import ArrayLike, NDArray
from sklearn.base import BaseEstimator, ClassifierMixin, MetaEstimatorMixin, _fit_context, check_is_fitted, clone
from sklearn.linear_model import HuberRegressor
from sklearn.utils._available_if import available_if
from sklearn.utils._metadata_requests import RequestMethod
from sklearn.utils._param_validation import HasMethods, Interval, StrOptions

from ..._common import Parameter
from ..._types import FloatArrayLike, FloatNDArray, ParameterConstraint
from ...metrics import Metric
from ...utils._sklearn_compat import Tags, _estimator_has, validate_data  # type: ignore[attr-defined]
from ._cs_mixin import CostSensitiveMixin

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
CostStr = Literal['tp_cost', 'tn_cost', 'fn_cost', 'fp_cost']
K = TypeVar('K')
V = TypeVar('V')


class RobustCSClassifier(ClassifierMixin, MetaEstimatorMixin, CostSensitiveMixin, BaseEstimator):  # type: ignore[misc]
    """
    Cost-sensitive classifier that is robust to outliers in the instance-dependent costs.

    The costs are adjusted by fitting an outlier estimator to the costs and imputing the costs for the outliers.
    Outliers are detected by the standardized residuals of the cost and the predicted cost.
    The costs passed to the cost-sensitive classifier are a combination of the original costs (not non-outliers) and
    the imputed predicted costs (for outliers).

    Read more in the :ref:`User Guide <robustcs>`.

    Parameters
    ----------
    estimator : Estimator
        The cost-sensitive classifier to fit.
        The estimator must take tp_cost, tn_cost, fn_cost, and fp_cost as keyword arguments in its fit method
        or should use :class:`~empulse.metrics.Metric` as their loss/criterion.

    outlier_estimator : Estimator, optional
        The outlier estimator to fit to the costs.

        If not provided, a :class:`sklearn:sklearn.linear_model.HuberRegressor` is used with default settings.
    outlier_threshold : float, default=2.5
        The threshold for the standardized residuals to detect outliers.
        If the absolute value of the standardized residual is greater than the threshold,
        the cost is an outlier and will be imputed with the predicted cost.

    detect_outliers_for : {'all', 'tp_cost', 'tn_cost', 'fn_cost', 'fp_cost', list}, default='all'
        The costs for which to detect outliers.
        By default, all instance-dependent costs are used for outlier detection.
        If a single cost is passed, only that cost is used for outlier detection.
        If a list of costs is passed, only those costs are used for outlier detection.
        tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

        .. note::
            This parameter is ignored if the underlying estimator
            uses :class:`~empulse.metrics.Metric` as its loss/criterion.
            Then all costs that are marked as outlier-sensitive in the metric loss
            are used for outlier detection.
            This can be done through the :meth: `~empulse.metrics.Metric.mark_outlier_sensitive` method.

    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    tn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true negatives. If ``float``, then all true negatives have the same cost.
        If array-like, then it is the cost of each true negative classification.
        Is overwritten if another `tn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.


    Attributes
    ----------
    estimator_ : Estimator
        The fitted cost-sensitive classifier.
    outlier_estimators_ : dict{str, Estimator or None}
        The fitted outlier estimators.
        If no outliers are detected for this cost, the value is None.
        The keys of the directory are 'tp_cost', 'tn_cost', 'fn_cost', and 'fp_cost'.
    costs_ : dict
        The imputed costs for the cost-sensitive classifier.

    Notes
    -----
    Constant costs are not used for outlier detection and imputation.

    Code adapted from [1]_.

    Examples
    --------

    .. code-block:: python

        import numpy as np
        from empulse.models import CSLogitClassifier, RobustCSClassifier
        from sklearn.datasets import make_classification

        X, y = make_classification()
        fn_cost = np.random.rand(y.size)  # instance-dependent cost
        fp_cost = 5  # constant cost

        model = RobustCSClassifier(CSLogitClassifier(C=0.1))
        model.fit(X, y, fn_cost=fn_cost, fp_cost=fp_cost)

    Example with Metric loss:

    .. code-block:: python

        import numpy as np
        import sympy as sp
        from empulse.metrics import Metric, Cost, CostMatrix
        from empulse.models import CSLogitClassifier, RobustCSClassifier
        from sklearn.datasets import make_classification

        X, y = make_classification()
        a, b = sp.symbols('a b')
        cost_loss = Metric(
            CostMatrix().add_fp_cost(a).add_fn_cost(b).mark_outlier_sensitive(a), Cost()
        )
        fn_cost = np.random.rand(y.size)

        model = RobustCSClassifier(CSLogitClassifier(loss=cost_loss))
        model.fit(X, y, a=np.random.rand(y.size), b=5)

    Example with passing instance-dependent costs through cross-validation:

    .. code-block:: python

        import numpy as np
        from empulse.models import CSBoostClassifier, RobustCSClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import cross_val_score
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        fn_cost = np.random.rand(y.size)
        fp_cost = 5

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            (
                'model',
                RobustCSClassifier(CSBoostClassifier()).set_fit_request(
                    fn_cost=True, fp_cost=True
                ),
            ),
        ])

        cross_val_score(pipeline, X, y, params={'fn_cost': fn_cost, 'fp_cost': fp_cost})

    Example with passing instance-dependent costs through a grid search:

    .. code-block:: python

        import numpy as np
        from empulse.metrics import expected_cost_loss
        from empulse.models import CSLogitClassifier, RobustCSClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import GridSearchCV
        from sklearn.metrics import make_scorer
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler

        set_config(enable_metadata_routing=True)

        X, y = make_classification(n_samples=50)
        fn_cost = np.random.rand(y.size)
        fp_cost = 5

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            (
                'model',
                RobustCSClassifier(CSLogitClassifier()).set_fit_request(
                    fn_cost=True, fp_cost=True
                ),
            ),
        ])
        param_grid = {'model__estimator__C': np.logspace(-5, 2, 5)}
        scorer = make_scorer(
            expected_cost_loss,
            response_method='predict_proba',
            greater_is_better=False,
            normalize=True,
        )
        scorer = scorer.set_score_request(fn_cost=True, fp_cost=True)

        grid_search = GridSearchCV(pipeline, param_grid=param_grid, scoring=scorer)
        grid_search.fit(X, y, fn_cost=fn_cost, fp_cost=fp_cost)

    References
    ----------
    .. [1] De Vos, S., Vanderschueren, T., Verdonck, T., & Verbeke, W. (2023).
           Robust instance-dependent cost-sensitive classification.
           Advances in Data Analysis and Classification, 17(4), 1057-1079.
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'estimator': [HasMethods(['fit', 'predict_proba']), None],
        'outlier_estimator': [HasMethods(['fit', 'predict']), None],
        'outlier_threshold': [Interval(Real, 0, None, closed='right')],
        'detect_outliers_for': [StrOptions({'all', 'tp_cost', 'tn_cost', 'fn_cost', 'fp_cost'}), list],
        'tp_cost': ['array-like', Real],
        'tn_cost': ['array-like', Real],
        'fn_cost': ['array-like', Real],
        'fp_cost': ['array-like', Real],
    }

    def _get_metric_loss(self) -> Metric | None:
        """Get the metric loss function if available."""
        return self.estimator._get_metric_loss() if isinstance(self.estimator, CostSensitiveMixin) else None

    def __init__(
        self,
        estimator: Any,
        outlier_estimator: Any = None,
        *,
        outlier_threshold: float = 2.5,
        detect_outliers_for: Literal['all'] | CostStr | list[CostStr] = 'all',
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
    ):
        self.estimator = estimator
        self.outlier_estimator = outlier_estimator
        self.outlier_threshold = outlier_threshold
        self.detect_outliers_for = detect_outliers_for
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fn_cost = fn_cost
        self.fp_cost = fp_cost
        super().__init__()

    def __post_init__(self) -> None:
        # Allow passing costs accepted by the metric loss through metadata routing
        if isinstance(self._get_metric_loss(), Metric):
            self.__class__.set_fit_request = RequestMethod(
                'fit',
                sorted(
                    self.__class__._get_default_requests().fit.requests.keys() | self._get_metric_loss()._all_symbols  # type: ignore[union-attr]
                ),
            )

    @_fit_context(prefer_skip_nested_validation=False)  # type: ignore[misc]
    def fit(
        self,
        X: FloatArrayLike,
        y: ArrayLike,
        *,
        tp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        tn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        **fit_params: Any,
    ) -> Self:
        """
        Fit the estimator with the adjusted costs.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        y : array-like of shape (n_samples,)

        tp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true positives. If ``float``, then all true positives have the same cost.
            If array-like, then it is the cost of each true positive classification.

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        tn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true negatives. If ``float``, then all true negatives have the same cost.
            If array-like, then it is the cost of each true negative classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        fit_params : dict
            Additional keyword arguments to pass to the estimator's fit method.

        Returns
        -------
        self : RobustCSLogitClassifier
            Fitted RobustCSLogitClassifier model.
        """
        X, y = validate_data(self, X, y)

        if (
            isinstance(self.estimator, CostSensitiveMixin)
            and (metric_loss := self.estimator._get_metric_loss()) is not None
        ):
            self.costs_: dict[str, int | float | FloatNDArray] = {}
            outlier_symbols = metric_loss.cost_matrix._outlier_sensitive_symbols
            imputed_costs = {}

            self.outlier_estimators_ = {}
            for symbol in outlier_symbols:
                target = fit_params.get(str(symbol))
                if target is None:
                    alias = _invert_dict(metric_loss.cost_matrix._aliases)[str(symbol)]
                    target = fit_params.get(alias)
                    if target is None:
                        raise ValueError(f"Cost '{symbol}' is not provided in fit_params.")
                if not isinstance(target, np.ndarray):
                    raise ValueError(f"Cost '{symbol}' is not an array. Cannot detect outliers for this cost.")
                pos_symbols = metric_loss.tp_cost.free_symbols | metric_loss.fn_cost.free_symbols
                neg_symbols = metric_loss.tn_cost.free_symbols | metric_loss.fp_cost.free_symbols
                if symbol in pos_symbols and symbol not in neg_symbols:
                    X_relevant, target_relevant = X[y > 0], target[y > 0]
                elif symbol in neg_symbols and symbol not in pos_symbols:
                    X_relevant, target_relevant = X[y == 0], target[y == 0]
                else:
                    X_relevant, target_relevant = X.copy(), target.copy()

                if X_relevant.size > 0:
                    outlier_estimator = clone(
                        self.outlier_estimator if self.outlier_estimator is not None else HuberRegressor()
                    ).fit(X_relevant, target_relevant)
                    cost_predictions = outlier_estimator.predict(X)
                    residuals = np.abs(target - cost_predictions)
                    std_residuals = residuals / st.sem(target)
                    outliers = std_residuals > self.outlier_threshold
                    fit_params[str(symbol)] = np.where(outliers, cost_predictions, target)
                    self.costs_[str(symbol)] = fit_params[str(symbol)]
                    self.outlier_estimators_[str(symbol)] = outlier_estimator
        else:
            tp_cost, tn_cost, fn_cost, fp_cost = self._check_costs(
                tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost
            )
            self.costs_ = {
                'tp_cost': tp_cost if isinstance(tp_cost, int | float) else np.array(tp_cost),  # take copy of the array
                'tn_cost': tn_cost if isinstance(tn_cost, int | float) else np.array(tn_cost),
                'fn_cost': fn_cost if isinstance(fn_cost, int | float) else np.array(fn_cost),
                'fp_cost': fp_cost if isinstance(fp_cost, int | float) else np.array(fp_cost),
            }
            should_fit = self._determine_outlier_costs()
            self._fit_outlier_estimators(X, y, should_fit)
            imputed_costs = self.costs_.copy()

        # with the imputed costs fit the estimator
        self.estimator_ = clone(self.estimator).fit(X, y, **imputed_costs, **fit_params)

        if hasattr(self.estimator_, 'n_features_in_'):
            self.n_features_in_ = self.estimator_.n_features_in_
        if hasattr(self.estimator_, 'feature_names_in_'):
            self.feature_names_in_ = self.estimator_.feature_names_in_

        return self

    def _determine_outlier_costs(self) -> list[str]:
        """Determine which costs to fit the outlier estimator on."""
        # only fit on the costs that are arrays and have a standard deviation greater than 0
        should_fit: list[str] = [
            cost_name for cost_name, cost in self.costs_.items() if isinstance(cost, np.ndarray) and np.std(cost) > 0
        ]

        if self.detect_outliers_for != 'all':
            if isinstance(self.detect_outliers_for, str):
                if self.detect_outliers_for in self.costs_:  # single cost
                    if self.detect_outliers_for not in should_fit:
                        raise ValueError(
                            f"Cost '{self.detect_outliers_for}' is not an array or has a standard deviation of 0."
                            ' Cannot detect outliers for this cost.'
                        )
                    should_fit = [self.detect_outliers_for]  # type: ignore[list-item]
                else:
                    raise ValueError(
                        f"Invalid cost name '{self.detect_outliers_for}' in detect_outliers_for."
                        " Must be one of 'all', 'tp_cost', 'tn_cost', 'fn_cost', 'fp_cost', or a list of these."
                    )
            elif isinstance(self.detect_outliers_for, list):
                for cost_name in self.detect_outliers_for:
                    if cost_name not in self.costs_:
                        raise ValueError(f"Invalid cost name '{cost_name}' in detect_outliers_for.")
                    if cost_name not in should_fit:
                        raise ValueError(
                            f"Cost '{cost_name}' is not an array or has a standard deviation of 0."
                            ' Cannot detect outliers for this cost.'
                        )
                should_fit = [cost_name for cost_name in self.detect_outliers_for if cost_name in should_fit]
            else:
                raise TypeError(
                    f"Invalid type '{type(self.detect_outliers_for)}' for detect_outliers_for."
                    " Must be one of 'all', 'tp_cost', 'tn_cost', 'fn_cost', 'fp_cost', or a list of these."
                )

        return should_fit

    def _fit_outlier_estimators(self, X: FloatNDArray, y: FloatNDArray, should_fit: list[str]) -> None:
        self.outlier_estimators_ = {}
        for cost_name in self.costs_:
            if cost_name in should_fit:
                target = self.costs_[cost_name]
                if not isinstance(target, np.ndarray):
                    raise ValueError(f"Cost '{cost_name}' is not an array. Cannot detect outliers for this cost.")
                if cost_name in {'tp_cost', 'fn_cost'}:
                    X_relevant, target_relevant = X[y > 0], target[y > 0]
                else:
                    X_relevant, target_relevant = X[y == 0], target[y == 0]

                if X_relevant.size > 0:
                    outlier_estimator = clone(
                        self.outlier_estimator if self.outlier_estimator is not None else HuberRegressor()
                    ).fit(X_relevant, target_relevant)
                    cost_predictions = outlier_estimator.predict(X)
                    residuals = np.abs(target - cost_predictions)
                    std_residuals = residuals / st.sem(target)
                    outliers = std_residuals > self.outlier_threshold
                    self.costs_[cost_name] = np.where(outliers, cost_predictions, target)
                    self.outlier_estimators_[cost_name] = outlier_estimator
                else:
                    self.outlier_estimators_[cost_name] = None
            else:
                self.outlier_estimators_[cost_name] = None

    @available_if(_estimator_has('predict'))  # type: ignore[misc]
    def predict(self, X: FloatArrayLike) -> FloatNDArray:  # noqa: D102
        check_is_fitted(self, 'estimator_')
        y_pred: FloatNDArray = self.estimator_.predict(X)
        return y_pred

    @available_if(_estimator_has('predict_proba'))  # type: ignore[misc]
    def predict_proba(self, X: FloatArrayLike) -> FloatNDArray:  # noqa: D102
        check_is_fitted(self, 'estimator_')
        y_proba: FloatNDArray = self.estimator_.predict_proba(X)
        return y_proba

    @available_if(_estimator_has('decision_function'))  # type: ignore[misc]
    def decision_function(self, X: FloatArrayLike) -> FloatNDArray:  # noqa: D102
        check_is_fitted(self, 'estimator_')
        y_score: FloatNDArray = self.estimator_.decision_function(X)
        return y_score

    @property
    def classes_(self) -> NDArray[Any]:  # noqa: D102
        classes: NDArray[Any] = self.estimator_.classes_
        return classes

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags.multi_class = False
        tags.classifier_tags.poor_score = True
        return tags


def _invert_dict(d: MutableMapping[K, V]) -> dict[V, K]:
    """Invert a dictionary, swapping keys and values."""
    return {v: k for k, v in d.items()}



from functools import lru_cache

import numpy as np

from ..._types import FloatArrayLike, FloatNDArray
from ..common import _compute_profits
from ._validation import _validate_input_mp


def mpc_score(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    accept_rate: float = 0.3,
    clv: float = 200,
    incentive_cost: float = 10,
    contact_cost: float = 1,
    check_input: bool = True,
) -> float:
    """
    :func:`~empulse.metrics.mpc()` but only returning the MPC score.

    MPC presumes a situation where identified churners are contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer.
    For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.mpc_score` : to also return the fraction of the customer base
        that should be targeted to maximize profit.

        :func:`~empulse.metrics.empc_score` : for a stochastic version of this metric.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    accept_rate : float, default=0.3
        Probability of a customer accepting the retention offer (``0 < accept_rate < 1``).

    clv : float or 1D array-like, shape=(n_samples), default=200
        If ``float``: average customer lifetime value of retained customers (``clv > incentive_cost``).
        If ``array``: customer lifetime value of each customer when retained
        (``mean(clv) > incentive_cost``).

        .. note::
            Passing a CLV array is equivalent to passing a float with the average CLV of that array.

    incentive_cost : float, default=10
        Cost of incentive offered to a customer (``incentive_cost > 0``).

    contact_cost : float, default=1
        Cost of contacting a customer (``contact_cost > 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    mpc : float
        Maximum Profit Measure for Customer Churn.

    Notes
    -----
    The MPC is defined as [1]_:

    .. math::  CLV (\\gamma (1 - \\delta) - \\phi) \\pi_0 F_0(T) - CLV (\\delta + \\phi) \\pi_1 F_1(T)

    The MPC requires that the churn class is encoded as 0, and it is NOT interchangeable (see [2]_ p37).
    However, this implementation assumes the standard notation ('churn': 1, 'no churn': 0).

    An equivalent R implementation is available in [3]_.

    References
    ----------
    .. [1] Verbraken, T., Verbeke, W. and Baesens, B. (2013).
        A Novel Profit Maximizing Metric for Measuring Classification
        Performance of Customer Churn Prediction Models. IEEE Transactions on
        Knowledge and Data Engineering, 25(5), 961-973. Available Online:
        http://ieeexplore.ieee.org/iel5/69/6486492/06165289.pdf?arnumber=6165289
    .. [2] Bravo, C. and Vanden Broucke, S. and Verbraken, T. (2019).
        EMP: Expected Maximum Profit Classification Performance Measure.
        R package version 2.0.5. Available Online:
        http://cran.r-project.org/web/packages/EMP/index.html
    .. [3] Verbraken, T. (2013). Business-Oriented Data Analytics:
        Theory and Case Studies. Ph.D. dissertation, Dept. LIRIS, KU Leuven,
        Leuven, Belgium, 2013.

    Examples
    --------
    >>> from empulse.metrics import mpc_score
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> mpc_score(y_true, y_score)
    23.874999999999996

    Using scorer:

    >>> import numpy as np
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import cross_val_score, StratifiedKFold
    >>> from sklearn.metrics import make_scorer
    >>> from empulse.metrics import mpa_score
    >>>
    >>> X, y = make_classification(random_state=42)
    >>> model = LogisticRegression()
    >>> cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)
    >>> scorer = make_scorer(
    ...     mpc_score,
    ...     response_method='predict_proba',
    ...     clv=300,
    ...     incentive_cost=15,
    ... )
    >>> np.mean(cross_val_score(model, X, y, cv=cv, scoring=scorer))
    42.08999999999999
    """
    return mpc(
        y_true,
        y_score,
        clv=clv,
        incentive_cost=incentive_cost,
        contact_cost=contact_cost,
        accept_rate=accept_rate,
        check_input=check_input,
    )[0]


def mpc(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    accept_rate: float = 0.3,
    clv: FloatArrayLike | float = 200,
    incentive_cost: float = 10,
    contact_cost: float = 1,
    check_input: bool = True,
) -> tuple[float, float]:
    """
    Maximum Profit Measure for Customer Churn (MPC).

    MPC presumes a situation where identified churners are contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer.
    For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.mpc_score` : to only return the MPC score.

        :func:`~empulse.metrics.empc` : for a stochastic version of this metric.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    accept_rate : float, default=0.3
        Probability of a customer accepting the retention offer (``0 < accept_rate < 1``).

    clv : float or 1D array-like, shape=(n_samples), default=200
        If ``float``: average customer lifetime value of retained customers (``clv > incentive_cost``).
        If ``array``: customer lifetime value of each customer when retained
        (``mean(clv) > incentive_cost``).

        .. note::
            Passing a CLV array is equivalent to passing a float with the average CLV of that array.

    incentive_cost : float, default=10
        Cost of incentive offered to a customer (``incentive_cost > 0``).

    contact_cost : float, default=1
        Cost of contacting a customer (``contact_cost > 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empc : float
        Maximum Profit Measure for Customer Churn

    threshold : float
        Fraction of the customer base that should be targeted to maximize profit

    Notes
    -----
    The MPC is defined as [1]_:

    .. math::  CLV (\\gamma (1 - \\delta) - \\phi) \\pi_0 F_0(T) - CLV (\\delta + \\phi) \\pi_1 F_1(T)

    The MPC requires that the churn class is encoded as 0, and it is NOT interchangeable (see [3]_ p37).
    However, this implementation assumes the standard notation ('churn': 1, 'no churn': 0).

    An equivalent R implementation is available in [2]_.

    References
    ----------
    .. [1] Verbraken, T., Verbeke, W. and Baesens, B. (2013).
        A Novel Profit Maximizing Metric for Measuring Classification
        Performance of Customer Churn Prediction Models. IEEE Transactions on
        Knowledge and Data Engineering, 25(5), 961-973. Available Online:
        http://ieeexplore.ieee.org/iel5/69/6486492/06165289.pdf?arnumber=6165289
    .. [2] Bravo, C. and Vanden Broucke, S. and Verbraken, T. (2019).
        EMP: Expected Maximum Profit Classification Performance Measure.
        R package version 2.0.5. Available Online:
        http://cran.r-project.org/web/packages/EMP/index.html
    .. [3] Verbraken, T. (2013). Business-Oriented Data Analytics:
        Theory and Case Studies. Ph.D. dissertation, Dept. LIRIS, KU Leuven,
        Leuven, Belgium, 2013.

    Examples
    --------
    >>> from empulse.metrics import mpc
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> mpc(y_true, y_score)
    (23.874999999999996, 0.875)
    """
    profits, customer_thresholds = _compute_profit_churn(
        y_true,
        y_score,
        clv,
        incentive_cost,
        contact_cost,
        accept_rate,
        check_input,
    )
    max_profit_index = np.argmax(profits)

    return profits[max_profit_index], customer_thresholds[max_profit_index]


def _compute_profit_churn(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    clv: FloatArrayLike | float = 200,
    d: float = 10,
    f: float = 1,
    gamma: float = 0.3,
    check_input: bool = True,
) -> tuple[FloatNDArray, FloatNDArray]:
    if check_input:
        y_true, y_score, clv = _validate_input_mp(y_true, y_score, gamma, clv, d, f)
    else:
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
        clv = np.asarray(clv)
    if isinstance(clv, np.ndarray):
        clv = np.mean(clv)
    cost_benefits = _compute_cost_benefits(gamma, clv, d, f)
    return _compute_profits(y_true, y_score, cost_benefits)


@lru_cache(maxsize=1)
def _compute_cost_benefits(gamma: float, clv: float, d: float, f: float) -> FloatNDArray:
    delta = d / clv
    phi = f / clv

    true_positive_benefit = clv * (gamma * (1 - delta) - phi)
    false_positive_cost = -1 * clv * (delta + phi)
    return np.array([true_positive_benefit, false_positive_cost])
